# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from copy import deepcopy
from functools import partial
from typing import Optional, Sequence

import torch
from mmengine.hooks import Hook
from mmengine.model.wrappers import is_model_wrapper
from mmengine.registry import HOOKS
from mmengine.runner import Runner
from mmengine.utils import is_tuple_of

DATA_BATCH = Optional[Sequence[dict]]


@HOOKS.register_module()
class ExponentialMovingAverageHook(Hook):
    """Exponential Moving Average Hook.

    Exponential moving average is a trick that widely used in current GAN
    literature, e.g., PGGAN, StyleGAN, and BigGAN. This general idea of it is
    maintaining a model with the same architecture, but its parameters are
    updated as a moving average of the trained weights in the original model.
    In general, the model with moving averaged weights achieves better
    performance.

    Args:
        module_keys (str | tuple[str]): The name of the ema model. Note that we
            require these keys are followed by '_ema' so that we can easily
            find the original model by discarding the last four characters.
        interp_mode (str, optional): Mode of the interpolation method.
            Defaults to 'lerp'.
        interp_cfg (dict | None, optional): Set arguments of the interpolation
            function. Defaults to None.
        interval (int, optional): Evaluation interval (by iterations).
            Default: -1.
        start_iter (int, optional): Start iteration for ema. If the start
            iteration is not reached, the weights of ema model will maintain
            the same as the original one. Otherwise, its parameters are updated
            as a moving average of the trained weights in the original model.
            Default: 0.
    """

    def __init__(self,
                 module_keys,
                 interp_mode='lerp',
                 interp_cfg=None,
                 interval=-1,
                 start_iter=0):
        super().__init__()
        assert isinstance(module_keys, str) or is_tuple_of(module_keys, str)
        self.module_keys = (module_keys, ) if isinstance(module_keys,
                                                         str) else module_keys
        # sanity check for the format of module keys
        for k in self.module_keys:
            assert k.endswith(
                '_ema'), 'You should give keys that end with "_ema".'
        self.interp_mode = interp_mode
        self.interp_cfg = dict() if interp_cfg is None else deepcopy(
            interp_cfg)
        self.interval = interval
        self.start_iter = start_iter

        assert hasattr(
            self, interp_mode
        ), f'Currently, we do not support {self.interp_mode} for EMA.'
        self.interp_func = partial(
            getattr(self, interp_mode), **self.interp_cfg)

    @staticmethod
    def lerp(a, b, momentum=0.001, momentum_nontrainable=1., trainable=True):
        """Does a linear interpolation of two parameters/ buffers.

        Args:
            a (torch.Tensor): Interpolation start point, refer to orig state.
            b (torch.Tensor): Interpolation end point, refer to ema state.
            momentum (float, optional): The weight for the interpolation
                formula. Defaults to 0.001.
            momentum_nontrainable (float, optional): The weight for the
                interpolation formula used for nontrainable parameters.
                Defaults to 1..
            trainable (bool, optional): Whether input parameters is trainable.
                If set to False, momentum_nontrainable will be used.
                Defaults to True.
        Returns:
            torch.Tensor: Interpolation result.
        """
        assert 0.0 < momentum < 1.0, 'momentum must be in range (0.0, 1.0)'\
                                     f'but got {momentum}'
        assert 0.0 < momentum_nontrainable <= 1.0, (
            'momentum_nontrainable must be in range (0.0, 1.0] but got '
            f'{momentum_nontrainable}')
        if momentum > 0.5:
            warnings.warn(
                'The value of momentum in EMA is usually a small number,'
                'which is different from the conventional notion of '
                f'momentum but got {momentum}. Please make sure the '
                f'value is correct.')
        m = momentum if trainable else momentum_nontrainable
        return b + (a - b) * m

    def every_n_iters(self, runner: Runner, n: int):
        """This is the function to perform every n iterations.

        Args:
            runner (Runner): runner used to drive the whole pipeline
            n (int): the number of iterations

        Returns:
            int: the latest iterations
        """
        if runner.iter < self.start_iter:
            return True
        return (runner.iter + 1 - self.start_iter) % n == 0 if n > 0 else False

    @torch.no_grad()
    def after_train_iter(self,
                         runner: Runner,
                         batch_idx: int,
                         data_batch: DATA_BATCH = None,
                         outputs: Optional[dict] = None) -> None:
        """This is the function to perform after each training iteration.

        Args:
            runner (Runner): runner to drive the pipeline
            batch_idx (int): the id of batch
            data_batch (DATA_BATCH, optional): data batch. Defaults to None.
            outputs (Optional[dict], optional): output. Defaults to None.
        """

        if not self.every_n_iters(runner, self.interval):
            return

        model = runner.model.module if is_model_wrapper(
            runner.model) else runner.model

        for key in self.module_keys:
            # get current ema states
            ema_net = getattr(model, key)
            states_ema = ema_net.state_dict(keep_vars=False)
            # get currently original states
            net = getattr(model, key[:-4])
            states_orig = net.state_dict(keep_vars=True)

            for k, v in states_orig.items():
                if runner.iter < self.start_iter:
                    states_ema[k].data.copy_(v.data)
                else:
                    states_ema[k] = self.interp_func(
                        v, states_ema[k], trainable=v.requires_grad).detach()
            ema_net.load_state_dict(states_ema, strict=True)

    def before_run(self, runner: Runner):
        """This is the function perform before each run.

        Args:
            runner (Runner): runner used to drive the whole pipeline

        Raises:
            RuntimeError: error message
        """
        model = runner.model.module if is_model_wrapper(
            runner.model) else runner.model
        # sanity check for ema model
        for k in self.module_keys:
            if not hasattr(model, k) and not hasattr(model, k[:-4]):
                raise RuntimeError(
                    f'Cannot find both {k[:-4]} and {k} network for EMA hook.')
            if not hasattr(model, k) and hasattr(model, k[:-4]):
                setattr(model, k, deepcopy(getattr(model, k[:-4])))
                warnings.warn(
                    f'We do not suggest construct and initialize EMA model {k}'
                    ' in hook. You may explicitly define it by yourself.')



# Copyright (c) OpenMMLab. All rights reserved.
import math
import numbers
import random
from typing import Dict

import cv2
import numpy as np
import torch
import torchvision.transforms as transforms
from mmcv.transforms import BaseTransform
from PIL import Image

from mmedit.registry import TRANSFORMS


class BinarizeImage(BaseTransform):
    """Binarize image.

    Args:
        keys (Sequence[str]): The images to be binarized.
        binary_thr (float): Threshold for binarization.
        a_min (int): Lower limits of pixel value.
        a_max (int): Upper limits of pixel value.
        dtype (np.dtype): Set the data type of the output. Default: np.uint8
    """

    def __init__(self, keys, binary_thr, a_min=0, a_max=1, dtype=np.uint8):

        self.keys = keys
        self.binary_thr = binary_thr
        self.a_min = a_min
        self.a_max = a_max
        self.dtype = dtype

    def _binarize(self, img):
        """Binarize image.

        Args:
            img (np.ndarray): Input image.

        Returns:
            img (np.ndarray): Output image.
        """

        # Binarize to 0/1
        img = (img[..., :] > self.binary_thr).astype(np.uint8)

        if self.a_min != 0 or self.a_max != 1 or self.dtype != np.uint8:
            img = img * (self.a_max - self.a_min) + self.a_min
            img = img.astype(self.dtype)

        return img

    def transform(self, results):
        """The transform function of BinarizeImage.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """

        for k in self.keys:
            results[k] = self._binarize(results[k])

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (
            f'(keys={self.keys}, binary_thr={self.binary_thr}, '
            f'a_min={self.a_min}, a_max={self.a_max}, dtype={self.dtype})')

        return repr_str


@TRANSFORMS.register_module()
class Clip(BaseTransform):
    """Clip the pixels.

    Modified keys are the attributes specified in "keys".

    Args:
        keys (list[str]): The keys whose values are clipped.
        a_min (int): Lower limits of pixel value.
        a_max (int): Upper limits of pixel value.
    """

    def __init__(self, keys, a_min=0, a_max=255):

        self.keys = keys
        self.a_min = a_min
        self.a_max = a_max

    def _clip(self, input_):
        """Clip the pixels.

        Args:
            input_ (Union[List, np.ndarray]): Pixels to clip.

        Returns:
            Union[List, np.ndarray]: Clipped pixels.
        """
        is_single_image = False
        if isinstance(input_, np.ndarray):
            is_single_image = True
            input_ = [input_]

        # clip
        input_ = [np.clip(v, self.a_min, self.a_max) for v in input_]

        if is_single_image:
            input_ = input_[0]

        return input_

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict with the values of the specified keys are rounded
                and clipped.
        """

        for key in self.keys:
            results[key] = self._clip(results[key])

        return results

    def __repr__(self):

        result = self.__class__.__name__
        result += f'(a_min={self.a_min}, a_max={self.a_max})'

        return result


@TRANSFORMS.register_module()
class ColorJitter(BaseTransform):
    """An interface for torch color jitter so that it can be invoked in
    mmediting pipeline.

    Randomly change the brightness, contrast and saturation of an image.
    Modified keys are the attributes specified in "keys".

    Required Keys:

    - [KEYS]

    Modified Keys:

    - [KEYS]

    Args:
        keys (list[str]): The images to be resized.
        channel_order (str): Order of channel, candidates are 'bgr' and 'rgb'.
            Default: 'rgb'.

    Notes:

        ``**kwards`` follows the args list of
        ``torchvision.transforms.ColorJitter``.

        brightness (float or tuple of float (min, max)): How much to jitter
            brightness. brightness_factor is chosen uniformly from
            [max(0, 1 - brightness), 1 + brightness] or the given [min, max].
            Should be non negative numbers.
        contrast (float or tuple of float (min, max)): How much to jitter
            contrast. contrast_factor is chosen uniformly from
            [max(0, 1 - contrast), 1 + contrast] or the given [min, max].
            Should be non negative numbers.
        saturation (float or tuple of float (min, max)): How much to jitter
            saturation. saturation_factor is chosen uniformly from
            [max(0, 1 - saturation), 1 + saturation] or the given [min, max].
            Should be non negative numbers.
        hue (float or tuple of float (min, max)): How much to jitter hue.
            hue_factor is chosen uniformly from [-hue, hue] or the given
            [min, max].
            Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    """

    def __init__(self, keys, channel_order='rgb', **kwargs):

        assert keys, 'Keys should not be empty.'
        assert 'to_rgb' not in kwargs, (
            '`to_rgb` is not support in ColorJitter, '
            "which is replaced by `channel_order` ('rgb' or 'bgr')")

        self.keys = keys
        self.channel_order = channel_order
        self._transform = transforms.ColorJitter(**kwargs)

    def _color_jitter(self, image, this_seed):
        """Color Jitter Function.

        Args:
            image (np.ndarray): Image.
            this_seed (int): Seed of torch.

        Returns:
            image (np.ndarray): The output image.
        """

        if self.channel_order.lower() == 'bgr':
            image = image[..., ::-1]

        image = Image.fromarray(image)
        torch.manual_seed(this_seed)
        image = self._transform(image)
        image = np.asarray(image)

        if self.channel_order.lower() == 'bgr':
            image = image[..., ::-1]

        return image

    def transform(self, results: Dict) -> Dict:
        """The transform function of ColorJitter.

        Args:
            results (dict): The result dict.

        Returns:
            dict: The result dict.
        """

        this_seed = random.randint(0, 2**32)

        for k in self.keys:
            if isinstance(results[k], list):
                results[k] = [
                    self._color_jitter(v, this_seed) for v in results[k]
                ]
            else:
                results[k] = self._color_jitter(results[k], this_seed)

        return results

    def __repr__(self):

        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys}, channel_order={self.channel_order}, '
                     f'brightness={self._transform.brightness}, '
                     f'contrast={self._transform.contrast}, '
                     f'saturation={self._transform.saturation}, '
                     f'hue={self._transform.hue})')

        return repr_str


@TRANSFORMS.register_module()
class RandomAffine(BaseTransform):
    """Apply random affine to input images.

    This class is adopted from
    https://github.com/pytorch/vision/blob/v0.5.0/torchvision/transforms/
    transforms.py#L1015
    It should be noted that in
    https://github.com/Yaoyi-Li/GCA-Matting/blob/master/dataloader/
    data_generator.py#L70
    random flip is added. See explanation of `flip_ratio` below.
    Required keys are the keys in attribute "keys", modified keys
    are keys in attribute "keys".

    Args:
        keys (Sequence[str]): The images to be affined.
        degrees (float | tuple[float]): Range of degrees to select from. If it
            is a float instead of a tuple like (min, max), the range of degrees
            will be (-degrees, +degrees). Set to 0 to deactivate rotations.
        translate (tuple, optional): Tuple of maximum absolute fraction for
            horizontal and vertical translations. For example translate=(a, b),
            then horizontal shift is randomly sampled in the range
            -img_width * a < dx < img_width * a and vertical shift is randomly
            sampled in the range -img_height * b < dy < img_height * b.
            Default: None.
        scale (tuple, optional): Scaling factor interval, e.g (a, b), then
            scale is randomly sampled from the range a <= scale <= b.
            Default: None.
        shear (float | tuple[float], optional): Range of shear degrees to
            select from. If shear is a float, a shear parallel to the x axis
            and a shear parallel to the y axis in the range (-shear, +shear)
            will be applied. Else if shear is a tuple of 2 values, a x-axis
            shear and a y-axis shear in (shear[0], shear[1]) will be applied.
            Default: None.
        flip_ratio (float, optional): Probability of the image being flipped.
            The flips in horizontal direction and vertical direction are
            independent. The image may be flipped in both directions.
            Default: None.
    """

    def __init__(self,
                 keys,
                 degrees,
                 translate=None,
                 scale=None,
                 shear=None,
                 flip_ratio=None):

        self.keys = keys
        if isinstance(degrees, numbers.Number):
            assert degrees >= 0, ('If degrees is a single number, '
                                  'it must be positive.')
            self.degrees = (-degrees, degrees)
        else:
            assert isinstance(degrees, tuple) and len(degrees) == 2, \
                'degrees should be a tuple and it must be of length 2.'
            self.degrees = degrees

        if translate is not None:
            assert isinstance(translate, tuple) and len(translate) == 2, \
                'translate should be a tuple and it must be of length 2.'
            for t in translate:
                assert 0.0 <= t <= 1.0, ('translation values should be '
                                         'between 0 and 1.')
        self.translate = translate

        if scale is not None:
            assert isinstance(scale, tuple) and len(scale) == 2, \
                'scale should be a tuple and it must be of length 2.'
            for s in scale:
                assert s > 0, 'scale values should be positive.'
        self.scale = scale

        if shear is not None:
            if isinstance(shear, numbers.Number):
                assert shear >= 0, ('If shear is a single number, '
                                    'it must be positive.')
                self.shear = (-shear, shear)
            else:
                assert isinstance(shear, tuple) and len(shear) == 2, \
                    'shear should be a tuple and it must be of length 2.'
                # X-Axis and Y-Axis shear with (min, max)
                self.shear = shear
        else:
            self.shear = shear

        if flip_ratio is not None:
            assert isinstance(flip_ratio,
                              float), 'flip_ratio should be a float.'
            self.flip_ratio = flip_ratio
        else:
            self.flip_ratio = 0

    @staticmethod
    def _get_params(degrees, translate, scale_ranges, shears, flip_ratio,
                    img_size):
        """Get parameters for affine transformation.

        Returns:
            paras (tuple): Params to be passed to the affine transformation.
        """

        angle = np.random.uniform(degrees[0], degrees[1])
        if translate is not None:
            max_dx = translate[0] * img_size[0]
            max_dy = translate[1] * img_size[1]
            translations = (np.round(np.random.uniform(-max_dx, max_dx)),
                            np.round(np.random.uniform(-max_dy, max_dy)))
        else:
            translations = (0, 0)

        if scale_ranges is not None:
            scale = (np.random.uniform(scale_ranges[0], scale_ranges[1]),
                     np.random.uniform(scale_ranges[0], scale_ranges[1]))
        else:
            scale = (1.0, 1.0)

        if shears is not None:
            shear = np.random.uniform(shears[0], shears[1])
        else:
            shear = 0.0

        # Because `flip` is used as a multiplier in line 479 and 480,
        # so -1 stands for flip and 1 stands for no flip. Thus `flip`
        # should be an 'inverse' flag as the result of the comparison.
        # See https://github.com/open-mmlab/mmediting/pull/799 for more detail
        flip = (np.random.rand(2) > flip_ratio).astype(np.int32) * 2 - 1

        return angle, translations, scale, shear, flip

    @staticmethod
    def _get_inverse_affine_matrix(center, angle, translate, scale, shear,
                                   flip):
        """Helper method to compute inverse matrix for affine transformation.

        As it is explained in PIL.Image.rotate, we need compute INVERSE of
        affine transformation matrix: M = T * C * RSS * C^-1 where
        T is translation matrix:
            [1, 0, tx | 0, 1, ty | 0, 0, 1];
        C is translation matrix to keep center:
            [1, 0, cx | 0, 1, cy | 0, 0, 1];
        RSS is rotation with scale and shear matrix.

        It is different from the original function in torchvision.
        1. The order are changed to flip -> scale -> rotation -> shear.
        2. x and y have different scale factors.
        RSS(shear, a, scale, f) =
            [ cos(a + shear)*scale_x*f -sin(a + shear)*scale_y     0]
            [ sin(a)*scale_x*f          cos(a)*scale_y             0]
            [     0                       0                        1]
        Thus, the inverse is M^-1 = C * RSS^-1 * C^-1 * T^-1.
        """

        angle = math.radians(angle)
        shear = math.radians(shear)
        scale_x = 1.0 / scale[0] * flip[0]
        scale_y = 1.0 / scale[1] * flip[1]

        # Inverted rotation matrix with scale and shear
        d = math.cos(angle + shear) * math.cos(angle) + math.sin(
            angle + shear) * math.sin(angle)
        matrix = [
            math.cos(angle) * scale_x,
            math.sin(angle + shear) * scale_x, 0, -math.sin(angle) * scale_y,
            math.cos(angle + shear) * scale_y, 0
        ]
        matrix = [m / d for m in matrix]

        # Apply inverse of translation and of center translation:
        # RSS^-1 * C^-1 * T^-1
        matrix[2] += matrix[0] * (-center[0] - translate[0]) + matrix[1] * (
            -center[1] - translate[1])
        matrix[5] += matrix[3] * (-center[0] - translate[0]) + matrix[4] * (
            -center[1] - translate[1])

        # Apply center translation: C * RSS^-1 * C^-1 * T^-1
        matrix[2] += center[0]
        matrix[5] += center[1]

        return matrix

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """

        h, w = results[self.keys[0]].shape[:2]
        # if image is too small, set degree to 0 to reduce introduced dark area
        if np.maximum(h, w) < 1024:
            params = self._get_params((0, 0), self.translate, self.scale,
                                      self.shear, self.flip_ratio, (h, w))
        else:
            params = self._get_params(self.degrees, self.translate, self.scale,
                                      self.shear, self.flip_ratio, (h, w))

        center = (w * 0.5 - 0.5, h * 0.5 - 0.5)
        M = self._get_inverse_affine_matrix(center, *params)
        M = np.array(M).reshape((2, 3))

        for key in self.keys:
            ori_ndim = results[key].ndim
            results[key] = cv2.warpAffine(
                results[key],
                M, (w, h),
                flags=cv2.INTER_NEAREST + cv2.WARP_INVERSE_MAP)

            if ori_ndim == 3 and results[key].ndim == 2:
                results[key] = results[key][..., None]

        return results

    def __repr__(self):

        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys}, degrees={self.degrees}, '
                     f'translate={self.translate}, scale={self.scale}, '
                     f'shear={self.shear}, flip_ratio={self.flip_ratio})')

        return repr_str


@TRANSFORMS.register_module()
class RandomMaskDilation(BaseTransform):
    """Randomly dilate binary masks.

    Args:
        keys (Sequence[str]): The images to be resized.
        binary_thr (float): Threshold for obtaining binary mask. Default: 0.
        kernel_min (int): Min size of dilation kernel. Default: 9.
        kernel_max (int): Max size of dilation kernel. Default: 49.
    """

    def __init__(self, keys, binary_thr=0., kernel_min=9, kernel_max=49):

        self.keys = keys
        self.kernel_min = kernel_min
        self.kernel_max = kernel_max
        self.binary_thr = binary_thr

    def _random_dilate(self, img):

        kernel_size = np.random.randint(self.kernel_min, self.kernel_max + 1)
        kernel = np.ones((kernel_size, kernel_size), dtype=np.uint8)
        dilate_kernel_size = kernel_size
        img_ = cv2.dilate(img, kernel, iterations=1)

        img_ = (img_ > self.binary_thr).astype(np.float32)

        return img_, dilate_kernel_size

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """

        for k in self.keys:
            results[k], d_kernel = self._random_dilate(results[k])
            if len(results[k].shape) == 2:
                results[k] = np.expand_dims(results[k], axis=2)
            results[k + '_dilate_kernel_size'] = d_kernel

        return results

    def __repr__(self):

        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys}, kernel_min={self.kernel_min}, '
                     f'kernel_max={self.kernel_max})')

        return repr_str


@TRANSFORMS.register_module()
class UnsharpMasking(BaseTransform):
    """Apply unsharp masking to an image or a sequence of images.

    Args:
        kernel_size (int): The kernel_size of the Gaussian kernel.
        sigma (float): The standard deviation of the Gaussian.
        weight (float): The weight of the "details" in the final output.
        threshold (float): Pixel differences larger than this value are
            regarded as "details".
        keys (list[str]): The keys whose values are processed.

    Added keys are "xxx_unsharp", where "xxx" are the attributes specified
    in "keys".
    """

    def __init__(self, kernel_size, sigma, weight, threshold, keys):

        if kernel_size % 2 == 0:
            raise ValueError('kernel_size must be an odd number, but '
                             f'got {kernel_size}.')

        self.kernel_size = kernel_size
        self.sigma = sigma
        self.weight = weight
        self.threshold = threshold
        self.keys = keys

        kernel = cv2.getGaussianKernel(kernel_size, sigma)
        self.kernel = np.matmul(kernel, kernel.transpose())

    def _unsharp_masking(self, imgs):
        """Unsharp masking function."""

        is_single_image = False
        if isinstance(imgs, np.ndarray):
            is_single_image = True
            imgs = [imgs]

        outputs = []
        for img in imgs:
            img = img.astype(np.float32)
            residue = img - cv2.filter2D(img, -1, self.kernel)
            mask = np.float32(np.abs(residue) > self.threshold)
            soft_mask = cv2.filter2D(mask, -1, self.kernel)
            sharpened = np.clip(img + self.weight * residue, 0, 255)

            outputs.append(soft_mask * sharpened + (1 - soft_mask) * img)

        if is_single_image:
            outputs = outputs[0]

        return outputs

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        for key in self.keys:
            results[f'{key}_unsharp'] = self._unsharp_masking(results[key])

        return results

    def __repr__(self):

        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys}, kernel_size={self.kernel_size}, '
                     f'sigma={self.sigma}, weight={self.weight}, '
                     f'threshold={self.threshold})')

        return repr_str



test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackEditInputs')
]

rain100h_data_root = 'data/Rain100H'
rain100h_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Rain100H', task_name='deraining'),
        data_root=rain100h_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
rain100h_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='Rain100H'),
    dict(type='SSIM', convert_to='Y', prefix='Rain100H'),
]

rain100l_data_root = 'data/Rain100L'
rain100l_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Rain100L', task_name='deraining'),
        data_root=rain100l_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
rain100l_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='Rain100L'),
    dict(type='SSIM', convert_to='Y', prefix='Rain100L'),
]

test100_data_root = 'data/Test100'
test100_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Test100', task_name='deraining'),
        data_root=test100_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
test100_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='Test100'),
    dict(type='SSIM', convert_to='Y', prefix='Test100'),
]

test1200_data_root = 'data/Test1200'
test1200_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Test1200', task_name='deraining'),
        data_root=test1200_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
test1200_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='Test1200'),
    dict(type='SSIM', convert_to='Y', prefix='Test1200'),
]

test2800_data_root = 'data/Test2800'
test2800_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='Test2800', task_name='deraining'),
        data_root=test2800_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
test2800_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='Test2800'),
    dict(type='SSIM', convert_to='Y', prefix='Test2800'),
]

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    rain100h_dataloader,
    rain100l_dataloader,
    test100_dataloader,
    test1200_dataloader,
    test2800_dataloader,
]
test_evaluator = [
    rain100h_evaluator,
    rain100l_evaluator,
    test100_evaluator,
    test1200_evaluator,
    test2800_evaluator,
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp
from itertools import cycle

import mmcv


def generate_json(data_root, seg_root, bg_root, all_data):
    """Generate training json list for Background Matting video dataset.

    Args:
        data_root (str): Background Matting video data root.
        seg_root (str): Segmentation of Background Matting video data root.
        bg_root (str): Background video data root.
        all_data (bool): Whether use the last 80 frames of each video. If True,
            the last 80 frames will be added to training json. In the original
            Background Matting github repo, due to the use of motion cue, the
            last 80 frames is not used.
    """
    # use fixed-camera data to train the Background Matting model
    video_root = osp.join(data_root, 'fixed-camera/train')
    if seg_root is None:
        seg_root = video_root
    if bg_root is None:
        bg_root = osp.join(data_root, 'background')

    video_dirs = [
        entry for entry in os.listdir(video_root)
        if osp.isdir(osp.join(video_root, entry))
    ]
    bg_dirs = [
        entry for entry in os.listdir(bg_root)
        if osp.isdir(osp.join(bg_root, entry))
    ]

    # create an iterator that loops over all the background video frames
    bg_frames = []
    for bg_dir in bg_dirs:
        bg_frames.extend([
            osp.join(bg_root, bg_dir, f)
            for f in sorted(mmcv.scandir(osp.join(bg_root, bg_dir)))
        ])
    bg_stream = cycle(bg_frames)

    data_infos = []

    for video_dir in video_dirs:
        video_full_path = osp.join(video_root, video_dir)
        seg_full_path = osp.join(seg_root, video_dir)
        num_frames = len(
            list(mmcv.scandir(video_full_path, suffix='_img.png')))
        # In the original Background Matting github repo, the
        # last 80 frames is not used.
        effective_frames = num_frames if all_data else num_frames - 80
        for i in range(1, effective_frames + 1):
            # Though it's not composited, to be consistent with adobe data,
            # we call the captured image `merged`.
            # Since background video may not be under the same directory as
            # the Background Matting video data, we use full path in Background
            # Matting dataset annotation file.
            merged = osp.join(video_full_path, f'{i:04d}_img.png')
            seg = osp.join(seg_full_path, f'{i:04d}_masksDL.png')
            bg = video_full_path + '.png'
            bg_sup = next(bg_stream)
            data_info = dict(
                merged_path=merged,
                seg_path=seg,
                bg_path=bg,
                bg_sup_path=bg_sup)
            data_infos.append(data_info)
    save_json_path = 'fixed_camera_train.json'
    mmcv.dump(data_infos, osp.join(data_root, save_json_path))


def parse_args():
    parser = argparse.ArgumentParser(
        description='Prepare Background Matting video dataset',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('data_root', help='Background Matting video data root')
    parser.add_argument(
        '--seg-root',
        help='Segmentation of Background Matting video data root. If not '
        'specified, it will be considered segmentation results are placed in '
        'the video frames folder just as the original repo.')
    parser.add_argument(
        '--bg-root',
        help='Background video data root. If not specified, it will use the '
        'three background videos in the Captured_Data folder.')
    parser.add_argument(
        '--all-data',
        action='store_true',
        help='Also use the last 80 frames of each video')
    return parser.parse_args()


def main():
    args = parse_args()
    if not osp.exists(args.data_root):
        raise FileNotFoundError(f'{args.data_root} does not exist!')

    print('generating Background Matting dataset annotation file...')
    generate_json(args.data_root, args.seg_root, args.bg_root, args.all_data)
    print('annotation file generated...')


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from .stylegan1 import StyleGAN1
from .stylegan1_discriminator import StyleGAN1Discriminator
from .stylegan1_generator import StyleGAN1Generator
from .stylegan1_modules import (Blur, ConstantInput, EqualLinearActModule,
                                NoiseInjection, make_kernel)
from .stylegan_utils import get_mean_latent, style_mixing

__all__ = [
    'Blur', 'ConstantInput', 'EqualLinearActModule', 'make_kernel',
    'StyleGAN1', 'StyleGAN1Discriminator', 'StyleGAN1Generator',
    'get_mean_latent', 'style_mixing', 'NoiseInjection'
]



# Copyright (c) OpenMMLab. All rights reserved.
import math

import torch
import torch.nn as nn
from mmengine.model import BaseModule

from mmedit.models.base_archs import PixelShufflePack, ResidualBlockNoBN
from mmedit.models.utils import make_layer
from mmedit.registry import MODELS


@MODELS.register_module()
class EDSRNet(BaseModule):
    """EDSR network structure.

    Paper: Enhanced Deep Residual Networks for Single Image Super-Resolution.
    Ref repo: https://github.com/thstkdgus35/EDSR-PyTorch

    Args:
        in_channels (int): Channel number of inputs.
        out_channels (int): Channel number of outputs.
        mid_channels (int): Channel number of intermediate features.
            Default: 64.
        num_blocks (int): Block number in the trunk network. Default: 16.
        upscale_factor (int): Upsampling factor. Support 2^n and 3.
            Default: 4.
        res_scale (float): Used to scale the residual in residual block.
            Default: 1.
        rgb_mean (list[float]): Image mean in RGB orders.
            Default: [0.4488, 0.4371, 0.4040], calculated from DIV2K dataset.
        rgb_std (list[float]): Image std in RGB orders. In EDSR, it uses
            [1.0, 1.0, 1.0]. Default: [1.0, 1.0, 1.0].
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 mid_channels=64,
                 num_blocks=16,
                 upscale_factor=4,
                 res_scale=1,
                 rgb_mean=[0.4488, 0.4371, 0.4040],
                 rgb_std=[1.0, 1.0, 1.0]):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.mid_channels = mid_channels
        self.num_blocks = num_blocks
        self.upscale_factor = upscale_factor

        self.mean = torch.Tensor(rgb_mean).view(1, -1, 1, 1)
        self.std = torch.Tensor(rgb_std).view(1, -1, 1, 1)

        self.conv_first = nn.Conv2d(in_channels, mid_channels, 3, padding=1)
        self.body = make_layer(
            ResidualBlockNoBN,
            num_blocks,
            mid_channels=mid_channels,
            res_scale=res_scale)
        self.conv_after_body = nn.Conv2d(mid_channels, mid_channels, 3, 1, 1)
        self.upsample = UpsampleModule(upscale_factor, mid_channels)
        self.conv_last = nn.Conv2d(
            mid_channels, out_channels, 3, 1, 1, bias=True)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """

        self.mean = self.mean.to(x)
        self.std = self.std.to(x)

        x = (x - self.mean) / self.std
        x = self.conv_first(x)
        res = self.conv_after_body(self.body(x))
        res += x

        x = self.conv_last(self.upsample(res))
        x = x * self.std + self.mean

        return x


class UpsampleModule(nn.Sequential):
    """Upsample module used in EDSR.

    Args:
        scale (int): Scale factor. Supported scales: 2^n and 3.
        mid_channels (int): Channel number of intermediate features.
    """

    def __init__(self, scale, mid_channels):
        modules = []
        if (scale & (scale - 1)) == 0:  # scale = 2^n
            for _ in range(int(math.log(scale, 2))):
                modules.append(
                    PixelShufflePack(
                        mid_channels, mid_channels, 2, upsample_kernel=3))
        elif scale == 3:
            modules.append(
                PixelShufflePack(
                    mid_channels, mid_channels, scale, upsample_kernel=3))
        else:
            raise ValueError(f'scale {scale} is not supported. '
                             'Supported scales: 2^n and 3.')

        super().__init__(*modules)



# Copyright (c) OpenMMLab. All rights reserved.

from .bbox_utils import extract_around_bbox, extract_bbox_patch
from .flow_warp import flow_warp
from .model_utils import (default_init_weights, generation_init_weights,
                          get_module_device, get_valid_noise_size,
                          get_valid_num_batches, make_layer, set_requires_grad)
from .sampling_utils import label_sample_fn, noise_sample_fn
from .tensor_utils import get_unknown_tensor, normalize_vecs

__all__ = [
    'default_init_weights', 'make_layer', 'flow_warp',
    'generation_init_weights', 'set_requires_grad', 'extract_bbox_patch',
    'extract_around_bbox', 'get_unknown_tensor', 'noise_sample_fn',
    'label_sample_fn', 'get_valid_num_batches', 'get_valid_noise_size',
    'get_module_device', 'normalize_vecs'
]



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Optional

from mmengine.dataset import pseudo_collate
from mmengine.runner import Runner
from torch.utils.data.dataloader import DataLoader


def _check_keys(sample_kwargs: dict, key: str) -> None:
    """Check whether target `key` is in the `sample_kwargs`."""
    assert key in sample_kwargs, (
        f'\'{key}\' must be set in \'sample_kwargs\'. But only find '
        f'following keys: \'{list(sample_kwargs.keys())}\'.')


def get_sampler(sample_kwargs: dict, runner: Optional[Runner]):
    """Get a sampler to loop input data.

    Args:
        sample_kwargs (dict): _description_
        runner (Optional[Runner]): _description_

    Returns:
        _type_: _description_
    """
    _check_keys(sample_kwargs, 'type')
    sampler_kwargs_ = deepcopy(sample_kwargs)
    sampler_type = sampler_kwargs_.pop('type')
    sampler = eval(f'{sampler_type}Sampler')(sampler_kwargs_, runner)
    return sampler


class ArgumentsSampler:
    """Dummy sampler only return input args multiple times."""

    def __init__(self,
                 sample_kwargs: dict,
                 runner: Optional[Runner] = None) -> None:
        _check_keys(sample_kwargs, 'max_times')
        assert isinstance(sample_kwargs['max_times'], int), (
            '\'max_times\' in \'sample_kwargs\' must be type of int.\'.')

        self.sample_kwargs = deepcopy(sample_kwargs)
        self.max_times = self.sample_kwargs.pop('max_times')
        self.forward_kwargs = self.sample_kwargs.pop('forward_kwargs')
        # set default num_batches from forward_kwargs
        self.forward_kwargs.setdefault('num_batches',
                                       self.sample_kwargs['num_batches'])
        self.idx = 0

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx >= self.max_times:
            raise StopIteration
        self.idx += 1
        return dict(inputs=deepcopy(self.forward_kwargs))


class NoiseSampler:
    """Noise sampler to by call `models.noise_fn` to generate noise."""

    def __init__(self, sample_kwargs: dict, runner: Runner) -> None:
        _check_keys(sample_kwargs, 'max_times')
        _check_keys(sample_kwargs, 'num_batches')

        self.sample_kwargs = deepcopy(sample_kwargs)
        self.max_times = self.sample_kwargs.pop('max_times')
        self.num_batches = self.sample_kwargs.pop('num_batches')

        module = runner.model
        if hasattr(module, 'module'):
            module = module.module
        self.module = module

        self.idx = 0

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx > self.max_times:
            raise StopIteration
        self.idx += 1

        noise = self.module.noise_fn(num_batches=self.num_batches)
        sample_kwargs = deepcopy(self.sample_kwargs)
        sample_kwargs['noise'] = noise
        # return sample_kwargs
        return dict(inputs=sample_kwargs)


class DataSampler:
    """Sampler loop the train_dataloader."""

    def __init__(self, sample_kwargs: dict, runner: Runner) -> None:
        _check_keys(sample_kwargs, 'max_times')

        self.sample_kwargs = deepcopy(sample_kwargs)
        self.max_times = self.sample_kwargs.pop('max_times')

        # build a new vanilla dataloader, because we should not reset the one
        # used in the training process.
        dataset = runner.train_dataloader.dataset
        batch_size = runner.train_dataloader.batch_size
        self._dataloader = DataLoader(
            dataset, batch_size=batch_size, collate_fn=pseudo_collate)
        self._iterator = iter(self._dataloader)

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx > self.max_times:
            self._iterator = iter(self._dataloader)
            raise StopIteration
        self.idx += 1
        return next(self._iterator)


class ValDataSampler:
    """Sampler loop the train_dataloader."""

    def __init__(self, sample_kwargs: dict, runner: Runner) -> None:
        _check_keys(sample_kwargs, 'max_times')

        self.sample_kwargs = deepcopy(sample_kwargs)
        self.max_times = self.sample_kwargs.pop('max_times')

        # build a new vanilla dataloader, because we should not reset the one
        # used in the training process.
        dataset = runner.val_dataloader.dataset
        batch_size = runner.val_dataloader.batch_size
        self._dataloader = DataLoader(
            dataset, batch_size=batch_size, collate_fn=pseudo_collate)
        self._iterator = iter(self._dataloader)

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx > self.max_times:
            self._iterator = iter(self._dataloader)
            raise StopIteration
        self.idx += 1
        return next(self._iterator)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from mmengine.optim import OptimWrapperDict

from mmedit.models.utils import set_requires_grad
from mmedit.registry import MODELS
from mmedit.structures import EditDataSample
from ..srgan import SRGAN


@MODELS.register_module()
class TTSR(SRGAN):
    """TTSR model for Reference-based Image Super-Resolution.

    Paper: Learning Texture Transformer Network for Image Super-Resolution.

    Args:
        generator (dict): Config for the generator.
        extractor (dict): Config for the extractor.
        transformer (dict): Config for the transformer.
        pixel_loss (dict): Config for the pixel loss.
        discriminator (dict): Config for the discriminator. Default: None.
        perceptual_loss (dict): Config for the perceptual loss. Default: None.
        transferal_perceptual_loss (dict): Config for the transferal perceptual
            loss. Default: None.
        gan_loss (dict): Config for the GAN loss. Default: None
        train_cfg (dict): Config for train. Default: None.
        test_cfg (dict): Config for testing. Default: None.
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Default: None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Default: None.
    """

    def __init__(self,
                 generator,
                 extractor,
                 transformer,
                 pixel_loss,
                 discriminator=None,
                 perceptual_loss=None,
                 transferal_perceptual_loss=None,
                 gan_loss=None,
                 train_cfg=None,
                 test_cfg=None,
                 init_cfg=None,
                 data_preprocessor=None):

        super().__init__(
            generator=generator,
            discriminator=discriminator,
            gan_loss=gan_loss,
            pixel_loss=pixel_loss,
            perceptual_loss=perceptual_loss,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            init_cfg=init_cfg,
            data_preprocessor=data_preprocessor)

        self.transformer = MODELS.build(transformer)
        self.extractor = MODELS.build(extractor)
        extractor['requires_grad'] = False
        self.extractor_copy = MODELS.build(extractor)
        set_requires_grad(self.extractor_copy, False)

        if transferal_perceptual_loss:
            self.transferal_perceptual_loss = MODELS.build(
                transferal_perceptual_loss)
        else:
            self.transferal_perceptual_loss = None

        self.pixel_init = train_cfg.get('pixel_init', 0) if train_cfg else 0

    def forward_tensor(self, inputs, data_samples=None, training=False):
        """Forward tensor. Returns result of simple forward.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.
            training (bool): Whether is training. Default: False.

        Returns:
            (Tensor | Tuple[List[Tensor]]): results of forward inference and
                forward train.
        """

        img_lq = []
        ref_lq = []
        ref = []

        img_lq = data_samples.img_lq / 255.
        ref_lq = data_samples.ref_lq / 255.
        ref = data_samples.ref_img / 255.

        img_lq, _, _ = self.extractor(img_lq)
        ref_lq, _, _ = self.extractor(ref_lq)
        refs = self.extractor(ref)

        soft_attention, textures = self.transformer(img_lq, ref_lq, refs)

        pred = self.generator(inputs, soft_attention, textures)

        if training:
            return pred, soft_attention, textures
        else:
            return pred

    def if_run_g(self):
        """Calculates whether need to run the generator step."""

        return True

    def if_run_d(self):
        """Calculates whether need to run the discriminator step."""

        return self.step_counter >= self.pixel_init and super().if_run_d()

    def g_step(self, batch_outputs, batch_gt_data: EditDataSample):
        """G step of GAN: Calculate losses of generator.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.

        Returns:
            dict: Dict of losses.
        """

        losses = dict()
        pred, soft_attention, textures = batch_outputs

        # pix loss
        if self.pixel_loss:
            losses['loss_pix'] = self.pixel_loss(pred, batch_gt_data)

        if self.step_counter >= self.pixel_init:
            # perceptual loss
            if self.perceptual_loss:
                loss_percep, loss_style = self.perceptual_loss(
                    pred, batch_gt_data)
                if loss_percep is not None:
                    losses['loss_perceptual'] = loss_percep
                if loss_style is not None:
                    losses['loss_style'] = loss_style

            # transform loss
            if self.transferal_perceptual_loss:
                state_dict = self.extractor.module.state_dict() if hasattr(
                    self.extractor, 'module') else self.extractor.state_dict()
                self.extractor_copy.load_state_dict(state_dict)
                sr_textures = self.extractor_copy((pred + 1.) / 2.)
                losses['loss_transferal'] = self.transferal_perceptual_loss(
                    sr_textures, soft_attention, textures)

            # gan loss for generator
            if self.gan_loss and self.discriminator:
                fake_g_pred = self.discriminator(pred)
                losses['loss_gan'] = self.gan_loss(
                    fake_g_pred, target_is_real=True, is_disc=False)

        return losses

    def g_step_with_optim(self, batch_outputs: torch.Tensor,
                          batch_gt_data: torch.Tensor,
                          optim_wrapper: OptimWrapperDict):
        """G step with optim of GAN: Calculate losses of generator and run
        optim.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.
            optim_wrapper (OptimWrapperDict): Optim wrapper dict.

        Returns:
            dict: Dict of parsed losses.
        """

        g_optim_wrapper = optim_wrapper['generator']
        e_optim_wrapper = optim_wrapper['extractor']

        losses_g = self.g_step(batch_outputs, batch_gt_data)
        parsed_losses_g, log_vars_g = self.parse_losses(losses_g)

        if g_optim_wrapper.should_update():
            g_optim_wrapper.backward(parsed_losses_g)
            g_optim_wrapper.step()
            g_optim_wrapper.zero_grad()

        if e_optim_wrapper.should_update():
            e_optim_wrapper.step()
            e_optim_wrapper.zero_grad()

        return log_vars_g

    def d_step_with_optim(self, batch_outputs, batch_gt_data, optim_wrapper):
        """D step with optim of GAN: Calculate losses of discriminator and run
        optim.

        Args:
            batch_outputs (Tuple[Tensor]): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.
            optim_wrapper (OptimWrapper): Optim wrapper of discriminator.

        Returns:
            dict: Dict of parsed losses.
        """

        pred, _, _ = batch_outputs

        return super().d_step_with_optim(
            batch_outputs=pred,
            batch_gt_data=batch_gt_data,
            optim_wrapper=optim_wrapper)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse

import mmcv
import torch

from mmedit.apis import init_model, inpainting_inference
from mmedit.utils import tensor2img


def parse_args():
    parser = argparse.ArgumentParser(description='Inpainting demo')
    parser.add_argument('config', help='test config file path')
    parser.add_argument('checkpoint', help='checkpoint file')
    parser.add_argument('masked_img_path', help='path to input image file')
    parser.add_argument('mask_path', help='path to input mask file')
    parser.add_argument('save_path', help='path to save inpainting result')
    parser.add_argument(
        '--imshow', action='store_true', help='whether show image with opencv')
    parser.add_argument('--device', type=int, default=0, help='CUDA device id')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    if args.device < 0 or not torch.cuda.is_available():
        device = torch.device('cpu')
    else:
        device = torch.device('cuda', args.device)

    model = init_model(args.config, args.checkpoint, device=device)

    result = inpainting_inference(model, args.masked_img_path, args.mask_path)
    result = tensor2img(result)[..., ::-1]

    mmcv.imwrite(result, args.save_path)
    if args.imshow:
        mmcv.imshow(result, 'predicted inpainting result')


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmcv.ops import ModulatedDeformConv2d, modulated_deform_conv2d
from mmengine.model import BaseModule
from mmengine.model.weight_init import constant_init, kaiming_init
from torch.nn.modules.utils import _pair

from mmedit.models.base_archs import PixelShufflePack, ResidualBlockNoBN
from mmedit.models.utils import make_layer
from mmedit.registry import MODELS


@MODELS.register_module()
class EDVRNet(BaseModule):
    """EDVR network structure for video super-resolution.

    Now only support X4 upsampling factor.
    Paper:
    EDVR: Video Restoration with Enhanced Deformable Convolutional Networks.

    Args:
        in_channels (int): Channel number of inputs.
        out_channels (int): Channel number of outputs.
        mid_channels (int): Channel number of intermediate features.
            Default: 64.
        num_frames (int): Number of input frames. Default: 5.
        deform_groups (int): Deformable groups. Defaults: 8.
        num_blocks_extraction (int): Number of blocks for feature extraction.
            Default: 5.
        num_blocks_reconstruction (int): Number of blocks for reconstruction.
            Default: 10.
        center_frame_idx (int): The index of center frame. Frame counting from
            0. Default: 2.
        with_tsa (bool): Whether to use TSA module. Default: True.
        init_cfg (dict, optional): Initialization config dict. Default: None.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 mid_channels=64,
                 num_frames=5,
                 deform_groups=8,
                 num_blocks_extraction=5,
                 num_blocks_reconstruction=10,
                 center_frame_idx=2,
                 with_tsa=True,
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)

        self.center_frame_idx = center_frame_idx
        self.with_tsa = with_tsa
        act_cfg = dict(type='LeakyReLU', negative_slope=0.1)

        self.conv_first = nn.Conv2d(in_channels, mid_channels, 3, 1, 1)
        self.feature_extraction = make_layer(
            ResidualBlockNoBN,
            num_blocks_extraction,
            mid_channels=mid_channels)

        # generate pyramid features
        self.feat_l2_conv1 = ConvModule(
            mid_channels, mid_channels, 3, 2, 1, act_cfg=act_cfg)
        self.feat_l2_conv2 = ConvModule(
            mid_channels, mid_channels, 3, 1, 1, act_cfg=act_cfg)
        self.feat_l3_conv1 = ConvModule(
            mid_channels, mid_channels, 3, 2, 1, act_cfg=act_cfg)
        self.feat_l3_conv2 = ConvModule(
            mid_channels, mid_channels, 3, 1, 1, act_cfg=act_cfg)
        # pcd alignment
        self.pcd_alignment = PCDAlignment(
            mid_channels=mid_channels, deform_groups=deform_groups)
        # fusion
        if self.with_tsa:
            self.fusion = TSAFusion(
                mid_channels=mid_channels,
                num_frames=num_frames,
                center_frame_idx=self.center_frame_idx)
        else:
            self.fusion = nn.Conv2d(num_frames * mid_channels, mid_channels, 1,
                                    1)

        # reconstruction
        self.reconstruction = make_layer(
            ResidualBlockNoBN,
            num_blocks_reconstruction,
            mid_channels=mid_channels)
        # upsample
        self.upsample1 = PixelShufflePack(
            mid_channels, mid_channels, 2, upsample_kernel=3)
        self.upsample2 = PixelShufflePack(
            mid_channels, 64, 2, upsample_kernel=3)
        # we fix the output channels in the last few layers to 64.
        self.conv_hr = nn.Conv2d(64, 64, 3, 1, 1)
        self.conv_last = nn.Conv2d(64, out_channels, 3, 1, 1)
        self.img_upsample = nn.Upsample(
            scale_factor=4, mode='bilinear', align_corners=False)
        # activation function
        self.lrelu = nn.LeakyReLU(negative_slope=0.1, inplace=True)

    def forward(self, x):
        """Forward function for EDVRNet.

        Args:
            x (Tensor): Input tensor with shape (n, t, c, h, w).

        Returns:
            Tensor: SR center frame with shape (n, c, h, w).
        """
        n, t, c, h, w = x.size()
        assert h % 4 == 0 and w % 4 == 0, (
            'The height and width of inputs should be a multiple of 4, '
            f'but got {h} and {w}.')

        x_center = x[:, self.center_frame_idx, :, :, :].contiguous()

        # extract LR features
        # L1
        l1_feat = self.lrelu(self.conv_first(x.view(-1, c, h, w)))
        l1_feat = self.feature_extraction(l1_feat)
        # L2
        l2_feat = self.feat_l2_conv2(self.feat_l2_conv1(l1_feat))
        # L3
        l3_feat = self.feat_l3_conv2(self.feat_l3_conv1(l2_feat))

        l1_feat = l1_feat.view(n, t, -1, h, w)
        l2_feat = l2_feat.view(n, t, -1, h // 2, w // 2)
        l3_feat = l3_feat.view(n, t, -1, h // 4, w // 4)

        # pcd alignment
        ref_feats = [  # reference feature list
            l1_feat[:, self.center_frame_idx, :, :, :].clone(),
            l2_feat[:, self.center_frame_idx, :, :, :].clone(),
            l3_feat[:, self.center_frame_idx, :, :, :].clone()
        ]
        aligned_feat = []
        for i in range(t):
            neighbor_feats = [
                l1_feat[:, i, :, :, :].clone(), l2_feat[:, i, :, :, :].clone(),
                l3_feat[:, i, :, :, :].clone()
            ]
            aligned_feat.append(self.pcd_alignment(neighbor_feats, ref_feats))
        aligned_feat = torch.stack(aligned_feat, dim=1)  # (n, t, c, h, w)

        if self.with_tsa:
            feat = self.fusion(aligned_feat)
        else:
            aligned_feat = aligned_feat.view(n, -1, h, w)
            feat = self.fusion(aligned_feat)

        # reconstruction
        out = self.reconstruction(feat)
        out = self.lrelu(self.upsample1(out))
        out = self.lrelu(self.upsample2(out))
        out = self.lrelu(self.conv_hr(out))
        out = self.conv_last(out)
        base = self.img_upsample(x_center)
        out += base
        return out

    def init_weights(self):
        """Init weights for models."""

        super().init_weights()

        init_type = None if self.init_cfg is None else self.init_cfg.get(
            'type', None)
        if init_type != 'Pretrained' and self.with_tsa:
            for module in [
                    self.fusion.feat_fusion, self.fusion.spatial_attn1,
                    self.fusion.spatial_attn2, self.fusion.spatial_attn3,
                    self.fusion.spatial_attn4, self.fusion.spatial_attn_l1,
                    self.fusion.spatial_attn_l2, self.fusion.spatial_attn_l3,
                    self.fusion.spatial_attn_add1
            ]:
                kaiming_init(
                    module.conv,
                    a=0.1,
                    mode='fan_out',
                    nonlinearity='leaky_relu',
                    bias=0,
                    distribution='uniform')


class ModulatedDCNPack(ModulatedDeformConv2d):
    """Modulated Deformable Convolutional Pack.

    Different from the official DCN, which generates offsets and masks from
    the preceding features, this ModulatedDCNPack takes another different
    feature to generate masks and offsets.

    Args:
        in_channels (int): Same as nn.Conv2d.
        out_channels (int): Same as nn.Conv2d.
        kernel_size (int or tuple[int]): Same as nn.Conv2d.
        stride (int or tuple[int]): Same as nn.Conv2d.
        padding (int or tuple[int]): Same as nn.Conv2d.
        dilation (int or tuple[int]): Same as nn.Conv2d.
        groups (int): Same as nn.Conv2d.
        bias (bool or str): If specified as `auto`, it will be decided by the
            norm_cfg. Bias will be set as True if norm_cfg is None, otherwise
            False.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.conv_offset = nn.Conv2d(
            self.in_channels,
            self.deform_groups * 3 * self.kernel_size[0] * self.kernel_size[1],
            kernel_size=self.kernel_size,
            stride=_pair(self.stride),
            padding=_pair(self.padding),
            bias=True)
        self.init_offset()

    def init_offset(self):
        """Init constant offset."""
        constant_init(self.conv_offset, val=0, bias=0)

    def forward(self, x, extra_feat):
        """Forward function."""
        out = self.conv_offset(extra_feat)
        o1, o2, mask = torch.chunk(out, 3, dim=1)
        offset = torch.cat((o1, o2), dim=1)
        mask = torch.sigmoid(mask)
        return modulated_deform_conv2d(x, offset, mask, self.weight, self.bias,
                                       self.stride, self.padding,
                                       self.dilation, self.groups,
                                       self.deform_groups)


class PCDAlignment(BaseModule):
    """Alignment module using Pyramid, Cascading and Deformable convolution
    (PCD). It is used in EDVRNet.

    Args:
        mid_channels (int): Number of the channels of middle features.
            Default: 64.
        deform_groups (int): Deformable groups. Defaults: 8.
        act_cfg (dict): Activation function config for ConvModule.
            Default: LeakyReLU with negative_slope=0.1.
    """

    def __init__(self,
                 mid_channels=64,
                 deform_groups=8,
                 act_cfg=dict(type='LeakyReLU', negative_slope=0.1)):
        super().__init__()

        # Pyramid has three levels:
        # L3: level 3, 1/4 spatial size
        # L2: level 2, 1/2 spatial size
        # L1: level 1, original spatial size
        self.offset_conv1 = nn.ModuleDict()
        self.offset_conv2 = nn.ModuleDict()
        self.offset_conv3 = nn.ModuleDict()
        self.dcn_pack = nn.ModuleDict()
        self.feat_conv = nn.ModuleDict()
        for i in range(3, 0, -1):
            level = f'l{i}'
            self.offset_conv1[level] = ConvModule(
                mid_channels * 2, mid_channels, 3, padding=1, act_cfg=act_cfg)
            if i == 3:
                self.offset_conv2[level] = ConvModule(
                    mid_channels, mid_channels, 3, padding=1, act_cfg=act_cfg)
            else:
                self.offset_conv2[level] = ConvModule(
                    mid_channels * 2,
                    mid_channels,
                    3,
                    padding=1,
                    act_cfg=act_cfg)
                self.offset_conv3[level] = ConvModule(
                    mid_channels, mid_channels, 3, padding=1, act_cfg=act_cfg)
            self.dcn_pack[level] = ModulatedDCNPack(
                mid_channels,
                mid_channels,
                3,
                padding=1,
                deform_groups=deform_groups)

            if i < 3:
                act_cfg_ = act_cfg if i == 2 else None
                self.feat_conv[level] = ConvModule(
                    mid_channels * 2,
                    mid_channels,
                    3,
                    padding=1,
                    act_cfg=act_cfg_)

        # Cascading DCN
        self.cas_offset_conv1 = ConvModule(
            mid_channels * 2, mid_channels, 3, padding=1, act_cfg=act_cfg)
        self.cas_offset_conv2 = ConvModule(
            mid_channels, mid_channels, 3, padding=1, act_cfg=act_cfg)
        self.cas_dcnpack = ModulatedDCNPack(
            mid_channels,
            mid_channels,
            3,
            padding=1,
            deform_groups=deform_groups)

        self.upsample = nn.Upsample(
            scale_factor=2, mode='bilinear', align_corners=False)
        self.lrelu = nn.LeakyReLU(negative_slope=0.1, inplace=True)

    def forward(self, neighbor_feats, ref_feats):
        """Forward function for PCDAlignment.

        Align neighboring frames to the reference frame in the feature level.

        Args:
            neighbor_feats (list[Tensor]): List of neighboring features. It
                contains three pyramid levels (L1, L2, L3),
                each with shape (n, c, h, w).
            ref_feats (list[Tensor]): List of reference features. It
                contains three pyramid levels (L1, L2, L3),
                each with shape (n, c, h, w).

        Returns:
            Tensor: Aligned features.
        """
        # The number of pyramid levels is 3.
        assert len(neighbor_feats) == 3 and len(ref_feats) == 3, (
            'The length of neighbor_feats and ref_feats must be both 3, '
            f'but got {len(neighbor_feats)} and {len(ref_feats)}')

        # Pyramids
        upsampled_offset, upsampled_feat = None, None
        for i in range(3, 0, -1):
            level = f'l{i}'
            offset = torch.cat([neighbor_feats[i - 1], ref_feats[i - 1]],
                               dim=1)
            offset = self.offset_conv1[level](offset)
            if i == 3:
                offset = self.offset_conv2[level](offset)
            else:
                offset = self.offset_conv2[level](
                    torch.cat([offset, upsampled_offset], dim=1))
                offset = self.offset_conv3[level](offset)

            feat = self.dcn_pack[level](neighbor_feats[i - 1], offset)
            if i == 3:
                feat = self.lrelu(feat)
            else:
                feat = self.feat_conv[level](
                    torch.cat([feat, upsampled_feat], dim=1))

            if i > 1:
                # upsample offset and features
                upsampled_offset = self.upsample(offset) * 2
                upsampled_feat = self.upsample(feat)

        # Cascading
        offset = torch.cat([feat, ref_feats[0]], dim=1)
        offset = self.cas_offset_conv2(self.cas_offset_conv1(offset))
        feat = self.lrelu(self.cas_dcnpack(feat, offset))
        return feat


class TSAFusion(BaseModule):
    """Temporal Spatial Attention (TSA) fusion module. It is used in EDVRNet.

    Args:
        mid_channels (int): Number of the channels of middle features.
            Default: 64.
        num_frames (int): Number of frames. Default: 5.
        center_frame_idx (int): The index of center frame. Default: 2.
        act_cfg (dict): Activation function config for ConvModule.
            Default: LeakyReLU with negative_slope=0.1.
    """

    def __init__(self,
                 mid_channels=64,
                 num_frames=5,
                 center_frame_idx=2,
                 act_cfg=dict(type='LeakyReLU', negative_slope=0.1)):
        super().__init__()
        self.center_frame_idx = center_frame_idx
        # temporal attention (before fusion conv)
        self.temporal_attn1 = nn.Conv2d(
            mid_channels, mid_channels, 3, padding=1)
        self.temporal_attn2 = nn.Conv2d(
            mid_channels, mid_channels, 3, padding=1)
        self.feat_fusion = ConvModule(
            num_frames * mid_channels, mid_channels, 1, act_cfg=act_cfg)

        # spatial attention (after fusion conv)
        self.max_pool = nn.MaxPool2d(3, stride=2, padding=1)
        self.avg_pool = nn.AvgPool2d(3, stride=2, padding=1)
        self.spatial_attn1 = ConvModule(
            num_frames * mid_channels, mid_channels, 1, act_cfg=act_cfg)
        self.spatial_attn2 = ConvModule(
            mid_channels * 2, mid_channels, 1, act_cfg=act_cfg)
        self.spatial_attn3 = ConvModule(
            mid_channels, mid_channels, 3, padding=1, act_cfg=act_cfg)
        self.spatial_attn4 = ConvModule(
            mid_channels, mid_channels, 1, act_cfg=act_cfg)
        self.spatial_attn5 = nn.Conv2d(
            mid_channels, mid_channels, 3, padding=1)
        self.spatial_attn_l1 = ConvModule(
            mid_channels, mid_channels, 1, act_cfg=act_cfg)
        self.spatial_attn_l2 = ConvModule(
            mid_channels * 2, mid_channels, 3, padding=1, act_cfg=act_cfg)
        self.spatial_attn_l3 = ConvModule(
            mid_channels, mid_channels, 3, padding=1, act_cfg=act_cfg)
        self.spatial_attn_add1 = ConvModule(
            mid_channels, mid_channels, 1, act_cfg=act_cfg)
        self.spatial_attn_add2 = nn.Conv2d(mid_channels, mid_channels, 1)

        self.lrelu = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.upsample = nn.Upsample(
            scale_factor=2, mode='bilinear', align_corners=False)

    def forward(self, aligned_feat):
        """Forward function for TSAFusion.

        Args:
            aligned_feat (Tensor): Aligned features with shape (n, t, c, h, w).

        Returns:
            Tensor: Features after TSA with the shape (n, c, h, w).
        """
        n, t, c, h, w = aligned_feat.size()
        # temporal attention
        embedding_ref = self.temporal_attn1(
            aligned_feat[:, self.center_frame_idx, :, :, :].clone())
        emb = self.temporal_attn2(aligned_feat.view(-1, c, h, w))
        emb = emb.view(n, t, -1, h, w)  # (n, t, c, h, w)

        corr_l = []  # correlation list
        for i in range(t):
            emb_neighbor = emb[:, i, :, :, :]
            corr = torch.sum(emb_neighbor * embedding_ref, 1)  # (n, h, w)
            corr_l.append(corr.unsqueeze(1))  # (n, 1, h, w)
        corr_prob = torch.sigmoid(torch.cat(corr_l, dim=1))  # (n, t, h, w)
        corr_prob = corr_prob.unsqueeze(2).expand(n, t, c, h, w)
        corr_prob = corr_prob.contiguous().view(n, -1, h, w)  # (n, t*c, h, w)
        aligned_feat = aligned_feat.view(n, -1, h, w) * corr_prob

        # fusion
        feat = self.feat_fusion(aligned_feat)

        # spatial attention
        attn = self.spatial_attn1(aligned_feat)
        attn_max = self.max_pool(attn)
        attn_avg = self.avg_pool(attn)
        attn = self.spatial_attn2(torch.cat([attn_max, attn_avg], dim=1))
        # pyramid levels
        attn_level = self.spatial_attn_l1(attn)
        attn_max = self.max_pool(attn_level)
        attn_avg = self.avg_pool(attn_level)
        attn_level = self.spatial_attn_l2(
            torch.cat([attn_max, attn_avg], dim=1))
        attn_level = self.spatial_attn_l3(attn_level)
        attn_level = self.upsample(attn_level)

        attn = self.spatial_attn3(attn) + attn_level
        attn = self.spatial_attn4(attn)
        attn = self.upsample(attn)
        attn = self.spatial_attn5(attn)
        attn_add = self.spatial_attn_add2(self.spatial_attn_add1(attn))
        attn = torch.sigmoid(attn)

        # after initialization, * 2 makes (attn * 2) to be close to 1.
        feat = feat * attn * 2 + attn_add
        return feat



_base_ = ['../singan/singan_balloons.py']

# TODO: have bugs
# MODEL
# NOTE: add by user, e.g.:
# test_pkl_data = './work_dirs/singan_pkl/singan_interp-pad_disc-nobn_balloons_20210406_180059-7d63e65d.pkl'  # noqa
test_pkl_data = None

model = dict(
    type='PESinGAN',
    generator=dict(
        type='SinGANMSGeneratorPE', interp_pad=True, noise_with_pad=True),
    discriminator=dict(norm_cfg=None),
    fixed_noise_with_pad=True)



# Copyright (c) OpenMMLab. All rights reserved.
import math

import clip
import lpips
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as T
import torchvision.transforms.functional as TF
from mmengine.utils import digit_version
from resize_right import resize
from torchvision import __version__ as TORCHVISION_VERSION

from mmedit.models.losses import tv_loss
from .secondary_model import alpha_sigma_to_t

normalize = T.Normalize(
    mean=[0.48145466, 0.4578275, 0.40821073],
    std=[0.26862954, 0.26130258, 0.27577711])


def sinc(x):
    """
    Sinc function.
    If x equal to 0,
        sinc(x) = 1
    else:
        sinc(x) = sin(x)/ x
    Args:
        x (torch.Tensor): Input Tensor

    Returns:
        torch.Tensor: Function output.
    """
    return torch.where(x != 0,
                       torch.sin(math.pi * x) / (math.pi * x), x.new_ones([]))


def lanczos(x, a):
    """Lanczos filter's reconstruction kernel L(x)."""
    cond = torch.logical_and(-a < x, x < a)
    out = torch.where(cond, sinc(x) * sinc(x / a), x.new_zeros([]))
    return out / out.sum()


def ramp(ratio, width):
    """_summary_

    Args:
        ratio (_type_): _description_
        width (_type_): _description_

    Returns:
        _type_: _description_
    """
    n = math.ceil(width / ratio + 1)
    out = torch.empty([n])
    cur = 0
    for i in range(out.shape[0]):
        out[i] = cur
        cur += ratio
    return torch.cat([-out[1:].flip([0]), out])[1:-1]


def resample(input, size, align_corners=True):
    """Lanczos resampling image.

    Args:
        input (torch.Tensor): Input image tensor.
        size (Tuple[int, int]): Output image size.
        align_corners (bool): align_corners argument of F.interpolate.
            Defaults to True.

    Returns:
        torch.Tensor: Resampling results.
    """
    n, c, h, w = input.shape
    dh, dw = size

    input = input.reshape([n * c, 1, h, w])

    if dh < h:
        kernel_h = lanczos(ramp(dh / h, 2), 2).to(input.device, input.dtype)
        pad_h = (kernel_h.shape[0] - 1) // 2
        input = F.pad(input, (0, 0, pad_h, pad_h), 'reflect')
        input = F.conv2d(input, kernel_h[None, None, :, None])

    if dw < w:
        kernel_w = lanczos(ramp(dw / w, 2), 2).to(input.device, input.dtype)
        pad_w = (kernel_w.shape[0] - 1) // 2
        input = F.pad(input, (pad_w, pad_w, 0, 0), 'reflect')
        input = F.conv2d(input, kernel_w[None, None, None, :])

    input = input.reshape([n, c, h, w])
    return F.interpolate(
        input, size, mode='bicubic', align_corners=align_corners)


def range_loss(input):
    """range loss."""
    return (input - input.clamp(-1, 1)).pow(2).mean([1, 2, 3])


def spherical_dist_loss(x, y):
    """spherical distance loss."""
    x = F.normalize(x, dim=-1)
    y = F.normalize(y, dim=-1)
    return (x - y).norm(dim=-1).div(2).arcsin().pow(2).mul(2)


class MakeCutouts(nn.Module):
    """Each iteration, the AI cuts the image into smaller pieces known as cuts.

    , and compares each cut to the prompt to decide how to guide the next
    diffusion step.
    This classes will randomly cut patches and perform image augmentation to
    these patches.

    Args:
        cut_size (int): Size of the patches.
        cutn (int): Number of patches to cut.
    """

    def __init__(self, cut_size, cutn):
        super().__init__()
        self.cut_size = cut_size
        self.cutn = cutn
        self.augs = T.Compose([
            T.RandomHorizontalFlip(p=0.5),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.RandomAffine(degrees=15, translate=(0.1, 0.1)),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.RandomPerspective(distortion_scale=0.4, p=0.7),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.RandomGrayscale(p=0.15),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
        ])

    def forward(self, input, skip_augs=False):
        input = T.Pad(input.shape[2] // 4, fill=0)(input)
        sideY, sideX = input.shape[2:4]
        max_size = min(sideX, sideY)

        cutouts = []
        for ch in range(self.cutn):
            if ch > self.cutn - self.cutn // 4:
                cutout = input.clone()
            else:
                size = int(max_size * torch.zeros(1, ).normal_(
                    mean=.8, std=.3).clip(float(self.cut_size / max_size), 1.))
                offsetx = torch.randint(0, abs(sideX - size + 1), ())
                offsety = torch.randint(0, abs(sideY - size + 1), ())
                cutout = input[:, :, offsety:offsety + size,
                               offsetx:offsetx + size]

            if not skip_augs:
                cutout = self.augs(cutout)
            cutouts.append(resample(cutout, (self.cut_size, self.cut_size)))
            del cutout

        cutouts = torch.cat(cutouts, dim=0)
        return cutouts


class MakeCutoutsDango(nn.Module):
    """Dango233(https://github.com/Dango233)'s version of MakeCutouts.

    The improvement compared to ``MakeCutouts`` is that it use partial
    greyscale augmentation to capture structure, and partial rotation
    augmentation to capture whole frames.

    Args:
        cut_size (int): Size of the patches.
        Overview (int): The total number of overview cuts.
        In details,
            Overview=1, Add whole frame;
            Overview=2, Add grayscaled frame;
            Overview=3, Add horizontal flip frame;
            Overview=4, Add grayscaled horizontal flip frame;
            Overview>4, Repeat add frame Overview times.
            Defaults to 4.
        InnerCrop (int): The total number of inner cuts.
            Defaults to 0.
        IC_Size_Pow (float): This sets the size of the border
            used for inner cuts.  High values have larger borders,
            and therefore the cuts themselves will be smaller and
            provide finer details. Defaults to 0.5.
        IC_Grey_P (float): The portion of the inner cuts can be set to be
            grayscale instead of color. This may help with improved
            definition of shapes and edges, especially in the early
            diffusion steps where the image structure is being defined.
            Defaults to 0.2.
    """

    def __init__(self,
                 cut_size,
                 Overview=4,
                 InnerCrop=0,
                 IC_Size_Pow=0.5,
                 IC_Grey_P=0.2):
        super().__init__()
        self.cut_size = cut_size
        self.Overview = Overview
        self.InnerCrop = InnerCrop
        self.IC_Size_Pow = IC_Size_Pow
        self.IC_Grey_P = IC_Grey_P

        random_affine_args = dict(degrees=10, translate=(0.05, 0.05))
        if digit_version(TORCHVISION_VERSION) >= digit_version('0.9.0'):
            random_affine_args['interpolation'] = T.InterpolationMode.BILINEAR
        else:
            from PIL import Image
            random_affine_args['resample'] = Image.NEAREST

        self.augs = T.Compose([
            T.RandomHorizontalFlip(p=0.5),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.RandomAffine(**random_affine_args),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.RandomGrayscale(p=0.1),
            T.Lambda(lambda x: x + torch.randn_like(x) * 0.01),
            T.ColorJitter(
                brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
        ])

    def forward(self, input, skip_augs=False):
        """Forward function."""
        cutouts = []
        gray = T.Grayscale(3)
        sideY, sideX = input.shape[2:4]
        max_size = min(sideX, sideY)
        min_size = min(sideX, sideY, self.cut_size)
        output_shape = [1, 3, self.cut_size, self.cut_size]
        pad_input = F.pad(input,
                          ((sideY - max_size) // 2, (sideY - max_size) // 2,
                           (sideX - max_size) // 2, (sideX - max_size) // 2))
        cutout = resize(pad_input, out_shape=output_shape)

        if self.Overview > 0:
            if self.Overview <= 4:
                if self.Overview >= 1:
                    cutouts.append(cutout)
                if self.Overview >= 2:
                    cutouts.append(gray(cutout))
                if self.Overview >= 3:
                    cutouts.append(TF.hflip(cutout))
                if self.Overview == 4:
                    cutouts.append(gray(TF.hflip(cutout)))
            else:
                cutout = resize(pad_input, out_shape=output_shape)
                for _ in range(self.Overview):
                    cutouts.append(cutout)

        if self.InnerCrop > 0:
            for i in range(self.InnerCrop):
                size = int(
                    torch.rand([])**self.IC_Size_Pow * (max_size - min_size) +
                    min_size)
                offsetx = torch.randint(0, sideX - size + 1, ())
                offsety = torch.randint(0, sideY - size + 1, ())
                cutout = input[:, :, offsety:offsety + size,
                               offsetx:offsetx + size]
                if i <= int(self.IC_Grey_P * self.InnerCrop):
                    cutout = gray(cutout)
                cutout = resize(cutout, out_shape=output_shape)
                cutouts.append(cutout)
        cutouts = torch.cat(cutouts)
        if not skip_augs:
            cutouts = self.augs(cutouts)
        return cutouts


def parse_prompt(prompt):
    """Parse prompt, return text and text weight."""
    if prompt.startswith('http://') or prompt.startswith('https://'):
        vals = prompt.rsplit(':', 2)
        vals = [vals[0] + ':' + vals[1], *vals[2:]]
    else:
        vals = prompt.rsplit(':', 1)
    vals = vals + ['', '1'][len(vals):]
    return vals[0], float(vals[1])


def split_prompts(prompts, max_frames=1):
    """Split prompts to a list of prompts."""
    prompt_series = pd.Series([np.nan for a in range(max_frames)])
    for i, prompt in prompts.items():
        prompt_series[i] = prompt
    # prompt_series = prompt_series.astype(str)
    prompt_series = prompt_series.ffill().bfill()
    return prompt_series


class ImageTextGuider(nn.Module):
    """Disco-Diffusion uses text and images to guide image generation. We will
    use the clip models to extract text and image features as prompts, and then
    during the iteration, the features of the image patches are computed, and
    the similarity loss between the prompts features and the generated features
    is computed. Other losses also include RGB Range loss, total variation
    loss. Using these losses we can guide the image generation towards the
    desired target.

    Args:
        clip_models (List[Dict]): List of clip model settings.
    """

    def __init__(self, clip_models):
        super().__init__()
        self.clip_models = clip_models
        self.lpips_model = lpips.LPIPS(net='vgg')

    def frame_prompt_from_text(self, text_prompts, frame_num=0):
        """Get current frame prompt."""
        prompts_series = split_prompts(text_prompts)
        if prompts_series is not None and frame_num >= len(prompts_series):
            frame_prompt = prompts_series[-1]
        elif prompts_series is not None:
            frame_prompt = prompts_series[frame_num]
        else:
            frame_prompt = []
        return frame_prompt

    def compute_prompt_stats(self,
                             text_prompts=[],
                             image_prompt=None,
                             fuzzy_prompt=False,
                             rand_mag=0.05):
        """Compute prompts statistics.

        Args:
            text_prompts (list): Text prompts. Defaults to [].
            image_prompt (list): Image prompts. Defaults to None.
            fuzzy_prompt (bool, optional): Controls whether to add multiple
                noisy prompts to the prompt losses. If True, can increase
                variability of image output. Defaults to False.
            rand_mag (float, optional): Controls the magnitude of the
                random noise added by fuzzy_prompt. Defaults to 0.05.
        """
        model_stats = []
        frame_prompt = self.frame_prompt_from_text(text_prompts)
        for clip_model in self.clip_models:
            model_stat = {
                'clip_model': None,
                'target_embeds': [],
                'make_cutouts': None,
                'weights': []
            }
            model_stat['clip_model'] = clip_model

            for prompt in frame_prompt:
                txt, weight = parse_prompt(prompt)
                txt = clip_model.model.encode_text(
                    clip.tokenize(prompt).to(self.device)).float()

                if fuzzy_prompt:
                    for i in range(25):
                        model_stat['target_embeds'].append(
                            (txt +
                             torch.randn(txt.shape).cuda() * rand_mag).clamp(
                                 0, 1))
                        model_stat['weights'].append(weight)
                else:
                    model_stat['target_embeds'].append(txt)
                    model_stat['weights'].append(weight)
            model_stat['target_embeds'] = torch.cat(
                model_stat['target_embeds'])
            model_stat['weights'] = torch.tensor(
                model_stat['weights'], device=self.device)
            if model_stat['weights'].sum().abs() < 1e-3:
                raise RuntimeError('The weights must not sum to 0.')
            model_stat['weights'] /= model_stat['weights'].sum().abs()
            model_stats.append(model_stat)
        return model_stats

    def cond_fn(self,
                model,
                diffusion_scheduler,
                x,
                t,
                beta_prod_t,
                model_stats,
                secondary_model=None,
                init_image=None,
                clamp_grad=True,
                clamp_max=0.05,
                clip_guidance_scale=5000,
                init_scale=1000,
                tv_scale=0.,
                sat_scale=0.,
                range_scale=150,
                cut_overview=[12] * 400 + [4] * 600,
                cut_innercut=[4] * 400 + [12] * 600,
                cut_ic_pow=[1] * 1000,
                cut_icgray_p=[0.2] * 400 + [0] * 600,
                cutn_batches=4):
        """Clip guidance function.

        Args:
            model (nn.Module): _description_
            diffusion_scheduler (object): _description_
            x (torch.Tensor): _description_
            t (int): _description_
            beta_prod_t (torch.Tensor): _description_
            model_stats (List[torch.Tensor]): _description_
            secondary_model (nn.Module): A smaller secondary diffusion model
                trained by Katherine Crowson to remove noise from intermediate
                timesteps to prepare them for CLIP.
                Ref: https://twitter.com/rivershavewings/status/1462859669454536711 # noqa
                Defaults to None.
            init_image (torch.Tensor): Initial image for denoising.
                Defaults to None.
            clamp_grad (bool, optional): Whether clamp gradient. Defaults to True.
            clamp_max (float, optional): Clamp max values. Defaults to 0.05.
            clip_guidance_scale (int, optional): The scale of influence of
                clip guidance on image generation. Defaults to 5000.
        """
        with torch.enable_grad():
            x_is_NaN = False
            x = x.detach().requires_grad_()
            n = x.shape[0]
            if secondary_model is not None:
                alpha = torch.tensor(
                    diffusion_scheduler.alphas_cumprod[t]**0.5,
                    dtype=torch.float32)
                sigma = torch.tensor(
                    (1 - diffusion_scheduler.alphas_cumprod[t])**0.5,
                    dtype=torch.float32)
                cosine_t = alpha_sigma_to_t(alpha, sigma).to(x.device)
                model_output = secondary_model(
                    x, cosine_t[None].repeat([x.shape[0]]))
                pred_original_sample = model_output['pred']
            else:
                model_output = model(x, t)['outputs']
                model_output, predicted_variance = torch.split(
                    model_output, x.shape[1], dim=1)
                alpha_prod_t = 1 - beta_prod_t
                pred_original_sample = (x - beta_prod_t**(0.5) *
                                        model_output) / alpha_prod_t**(0.5)
            # fac = diffusion_scheduler_output['beta_prod_t']** (0.5)
            # x_in = diffusion_scheduler_output['original_sample'] * fac + x * (1 - fac) # noqa
            fac = beta_prod_t**(0.5)
            x_in = pred_original_sample * fac + x * (1 - fac)
            x_in_grad = torch.zeros_like(x_in)
            for model_stat in model_stats:
                for i in range(cutn_batches):
                    t_int = int(t.item()) + 1
                    try:
                        input_resolution = model_stat[
                            'clip_model'].model.visual.input_resolution
                    except AttributeError:
                        input_resolution = 224

                    cuts = MakeCutoutsDango(
                        input_resolution,
                        Overview=cut_overview[1000 - t_int],
                        InnerCrop=cut_innercut[1000 - t_int],
                        IC_Size_Pow=cut_ic_pow[1000 - t_int],
                        IC_Grey_P=cut_icgray_p[1000 - t_int])
                    clip_in = normalize(cuts(x_in.add(1).div(2)))
                    image_embeds = model_stat['clip_model'].model.encode_image(
                        clip_in).float()
                    dists = spherical_dist_loss(
                        image_embeds.unsqueeze(1),
                        model_stat['target_embeds'].unsqueeze(0))
                    dists = dists.view([
                        cut_overview[1000 - t_int] +
                        cut_innercut[1000 - t_int], n, -1
                    ])
                    losses = dists.mul(model_stat['weights']).sum(2).mean(0)
                    x_in_grad += torch.autograd.grad(
                        losses.sum() * clip_guidance_scale,
                        x_in)[0] / cutn_batches
            tv_losses = tv_loss(x_in)
            range_losses = range_loss(pred_original_sample)
            sat_losses = torch.abs(x_in - x_in.clamp(min=-1, max=1)).mean()
            loss = tv_losses.sum() * tv_scale + range_losses.sum(
            ) * range_scale + sat_losses.sum() * sat_scale
            if init_image is not None and init_scale:
                init_losses = self.lpips_model(x_in, init_image)
                loss = loss + init_losses.sum() * init_scale
            x_in_grad += torch.autograd.grad(loss, x_in)[0]
            if not torch.isnan(x_in_grad).any():
                grad = -torch.autograd.grad(x_in, x, x_in_grad)[0]
            else:
                x_is_NaN = True
                grad = torch.zeros_like(x)
        if clamp_grad and not x_is_NaN:
            magnitude = grad.square().mean().sqrt()
            return grad * magnitude.clamp(max=clamp_max) / magnitude
        return grad

    @property
    def device(self):
        """Get current device of the model.

        Returns:
            torch.device: The current device of the model.
        """
        return next(self.parameters()).device

    def forward(self, x):
        """forward function."""
        raise NotImplementedError('No forward function for disco guider')



# Copyright (c) OpenMMLab. All rights reserved.
import json
import os
import os.path as osp
import subprocess
import traceback
import warnings
from typing import Dict, List, Optional, Union

import cv2
import gradio as gr
import numpy as np
import torch
import yaml
from mmengine.registry import init_default_scope

from mmedit.apis.inferencers.inpainting_inferencer import InpaintingInferencer


class InpaintingGradio:
    inpainting_supported_models = [
        # inpainting models
        'aot_gan',
        'deepfillv1',
        'deepfillv2',
        'global_local',
        'partial_conv',
    ]
    inpainting_supported_models_cfg = {}
    inpainting_supported_models_cfg_inited = False
    pkg_path = ''
    error_color = '#FF0000'
    success_color = '#00FF00'
    warning_color = '#FFFF00'
    notice_message = ('', None, '')

    def __init__(self,
                 model_name: str = None,
                 model_setting: int = None,
                 model_config: str = None,
                 model_ckpt: str = None,
                 device: torch.device = None,
                 extra_parameters: Dict = None,
                 seed: int = 2022,
                 **kwargs) -> None:
        init_default_scope('mmedit')
        InpaintingGradio.init_inference_supported_models_cfg()
        self.model_name = model_name
        self.model_setting = model_setting
        self.model_config = model_config
        self.model_ckpt = model_ckpt
        self.device = device
        self.extra_parameters = extra_parameters
        self.seed = seed
        if model_name or (model_config and model_ckpt):
            inpainting_kwargs = {}
            inpainting_kwargs.update(
                self._get_inpainting_kwargs(model_name, model_setting,
                                            model_config, model_ckpt,
                                            extra_parameters))
            self.inference = InpaintingInferencer(
                device=device, seed=seed, **inpainting_kwargs)

    def model_reconfig(self,
                       model_name: str = None,
                       model_setting: int = None,
                       model_config: str = None,
                       model_ckpt: str = None,
                       device: torch.device = None,
                       extra_parameters: Dict = None,
                       seed: int = 2022,
                       **kwargs) -> None:
        inpainting_kwargs = {}
        # if model_config:
        #     model_config = model_config.name
        # if model_ckpt:
        #     model_ckpt = model_ckpt.name
        if not model_name and model_setting:
            self.send_notification(
                'model_name should not be None when model_setting was used',
                self.error_color, 'error')
            return
        elif (not model_config and not model_name) and model_ckpt:
            self.send_notification(
                'model_name and model_config should not be None when '
                'model_ckpt was used', self.error_color, 'error')
            return
        elif (not model_ckpt and not model_name) and model_config:
            self.send_notification(
                'model_name and model_ckpt should not be None when '
                'model_config was used', self.error_color, 'error')
            return
        inpainting_kwargs.update(
            self._get_inpainting_kwargs(model_name, model_setting,
                                        model_config, model_ckpt,
                                        extra_parameters))
        try:
            self.inference = InpaintingInferencer(
                device=device, seed=seed, **inpainting_kwargs)
        except Exception as e:
            self.send_notification('inference Exception:' + str(e),
                                   self.error_color, 'error')
            traceback.print_exc()
            return
        self.send_notification('Model config Finished!', self.success_color,
                               'success')

    @staticmethod
    def change_text2dict(input_text: str) -> Union[Dict, None]:
        return_dict = None
        try:
            return_dict = json.loads(input_text)
        except Exception as e:
            InpaintingGradio.send_notification(
                'Convert string to dict Exception:' + str(e),
                InpaintingGradio.error_color, 'error')
        return return_dict

    @staticmethod
    def get_package_path() -> str:
        p = subprocess.Popen(
            'pip show mmedit', shell=True, stdout=subprocess.PIPE)
        out, err = p.communicate()
        out = out.decode()
        if 'Location' not in out:
            InpaintingGradio.send_notification('module mmedit not found',
                                               InpaintingGradio.error_color,
                                               'error')
            raise Exception('module mmedit not found')
        package_path = out[out.find('Location') +
                           len('Location: '):].split('\r\n')[0] + os.sep
        return package_path

    def get_model_config(self, model_name: str) -> Dict:
        """Get the model configuration including model config and checkpoint
        url.

        Args:
            model_name (str): Name of the model.
        Returns:
            dict: Model configuration.
        """
        if model_name not in self.inpainting_supported_models:
            self.send_notification(f'Model {model_name} is not supported.',
                                   self.error_color, 'error')
            raise ValueError(f'Model {model_name} is not supported.')
        else:
            return self.inpainting_supported_models_cfg[model_name]

    @staticmethod
    def init_inference_supported_models_cfg() -> None:
        if not InpaintingGradio.inpainting_supported_models_cfg_inited:
            InpaintingGradio.pkg_path = InpaintingGradio.get_package_path()
            # all_cfgs_dir = osp.join(osp.dirname(__file__), '..', 'configs')
            all_cfgs_dir = osp.join(InpaintingGradio.pkg_path, 'configs')
            for model_name in InpaintingGradio.inpainting_supported_models:
                meta_file_dir = osp.join(all_cfgs_dir, model_name,
                                         'metafile.yml')
                with open(meta_file_dir, 'r') as stream:
                    parsed_yaml = yaml.safe_load(stream)
                InpaintingGradio.inpainting_supported_models_cfg[
                    model_name] = {}
                InpaintingGradio.inpainting_supported_models_cfg[model_name][
                    'settings'] = parsed_yaml['Models']  # noqa
            InpaintingGradio.inpainting_supported_models_cfg_inited = True

    def _get_inpainting_kwargs(self, model_name: Optional[str],
                               model_setting: Optional[int],
                               model_config: Optional[str],
                               model_ckpt: Optional[str],
                               extra_parameters: Optional[Dict]) -> Dict:
        """Get the kwargs for the inpainting inferencer."""
        kwargs = {}

        if model_name:
            cfgs = self.get_model_config(model_name)
            # kwargs['task'] = cfgs['task']
            setting_to_use = 0
            if model_setting:
                if isinstance(model_setting, str):
                    model_setting = int(
                        model_setting[0:model_setting.find(' - ')])
                setting_to_use = model_setting
            else:
                model_setting = 0
            if model_setting > len(
                    cfgs['settings']) - 1 or model_setting < -len(
                        cfgs['settings']):
                self.send_notification(
                    f"model_setting out of range of {model_name}'s "
                    'cfgs settings', self.error_color, 'error')
            config_dir = cfgs['settings'][setting_to_use]['Config']
            config_dir = config_dir[config_dir.find('configs'):]
            # kwargs['config'] = os.path.join(
            #     osp.dirname(__file__), '..', config_dir)
            kwargs['config'] = os.path.join(self.pkg_path, config_dir)
            kwargs['ckpt'] = cfgs['settings'][setting_to_use]['Weights']

        if model_config:
            if kwargs.get('config', None) is not None:
                warnings.warn(
                    f'{model_name}\'s default config '
                    f'is overridden by {model_config}', UserWarning)
            kwargs['config'] = model_config

        if model_ckpt:
            if kwargs.get('ckpt', None) is not None:
                warnings.warn(
                    f'{model_name}\'s default checkpoint '
                    f'is overridden by {model_ckpt}', UserWarning)
            kwargs['ckpt'] = model_ckpt

        if extra_parameters:
            kwargs['extra_parameters'] = extra_parameters

        return kwargs

    @staticmethod
    def send_notification(msg: str, color: str, label: str) -> None:
        InpaintingGradio.notice_message = (msg, color, label)

    @staticmethod
    def get_inpainting_supported_models() -> List:
        """static function for getting inpainting inference supported modes."""
        return InpaintingGradio.inpainting_supported_models

    def infer(self, input_img_arg: Dict) -> np.ndarray:
        result = self.inference(
            img=input_img_arg['image'], mask=input_img_arg['mask'])
        result = cv2.cvtColor(result[1], cv2.COLOR_RGB2BGR)
        return result

    def run(self) -> None:
        with gr.Blocks() as demo:
            with gr.Row():
                with gr.Column():
                    input_model_dropdown = gr.Dropdown(
                        choices=self.inpainting_supported_models,
                        value=self.model_name,
                        label='choose model')
                    input_setting = gr.Dropdown(
                        value=self.model_setting, label='choose setting')
                    input_config = gr.Textbox(
                        value=self.model_config, label='model_config_path')
                    input_ckpt = gr.Textbox(
                        value=self.model_ckpt, label='model_ckpt_path')
                    input_device_dropdown = gr.Dropdown(
                        choices=['cuda', 'cpu'],
                        label='choose device',
                        value='cuda')
                    input_extra_parameters_input = gr.Textbox(
                        value=json.dumps(self.extra_parameters),
                        label='extra_parameters')
                    input_extra_parameters = gr.JSON(
                        value=self.extra_parameters, label='extra_parameters')
                    input_seed = gr.Number(
                        value=self.seed, precision=0, label='seed')
                    config_button = gr.Button('CONFIG')
                    input_extra_parameters_input.blur(
                        self.change_text2dict, input_extra_parameters_input,
                        input_extra_parameters)

                with gr.Column(visible=False) as output_col:
                    input_image = gr.Image(
                        image_mode='RGB',
                        tool='sketch',
                        type='filepath',
                        label='Input image')
                    infer_button = gr.Button('INFER')
                    infer_button.style(full_width=False)
                    output_image = gr.Image(
                        label='Output image', interactive=False)
                    output_image.style(height=500)
                    infer_button.click(
                        self.infer, inputs=input_image, outputs=output_image)

            with gr.Row():
                label = gr.Label(
                    value=self.notice_message[0],
                    color=self.notice_message[1],
                    label=self.notice_message[2])

            def show_infer(*args) -> Dict:
                self.model_reconfig(*args)
                return {
                    output_col:
                    gr.update(visible=True),
                    label:
                    gr.update(
                        value=self.notice_message[0],
                        color=self.notice_message[1],
                        label=self.notice_message[2])
                }

            def change_setting(model_name: str) -> Dict:
                settings = InpaintingGradio.inpainting_supported_models_cfg[
                    model_name]['settings']
                return {
                    input_setting:
                    gr.update(choices=[
                        str(i) + ' - ' + settings[i]['Config']
                        for i in range(len(settings))
                    ])
                }

            input_model_dropdown.change(
                fn=change_setting,
                inputs=input_model_dropdown,
                outputs=input_setting)

            config_button.click(show_infer, [
                input_model_dropdown,
                input_setting,
                input_config,
                input_ckpt,
                input_device_dropdown,
                input_extra_parameters,
                input_seed,
            ], [output_col, label])
        demo.launch()


if __name__ == '__main__':
    inpaintingGradio = InpaintingGradio()
    inpaintingGradio.run()



# DistributedDataParallel
model_wrapper_cfg = dict(type='MMSeparateDistributedDataParallel')

model = dict(
    type='PConvInpaintor',
    data_preprocessor=dict(
        type='EditDataPreprocessor',
        mean=[127.5],
        std=[127.5],
    ),
    encdec=dict(
        type='PConvEncoderDecoder',
        encoder=dict(
            type='PConvEncoder',
            norm_cfg=dict(type='SyncBN', requires_grad=False),
            norm_eval=True),
        decoder=dict(type='PConvDecoder', norm_cfg=dict(type='SyncBN'))),
    disc=None,
    loss_composed_percep=dict(
        type='PerceptualLoss',
        vgg_type='vgg16',
        layer_weights={
            '4': 1.,
            '9': 1.,
            '16': 1.,
        },
        perceptual_weight=0.05,
        style_weight=120,
        pretrained=('torchvision://vgg16')),
    loss_out_percep=True,
    loss_l1_hole=dict(
        type='L1Loss',
        loss_weight=6.,
    ),
    loss_l1_valid=dict(
        type='L1Loss',
        loss_weight=1.,
    ),
    loss_tv=dict(
        type='MaskedTVLoss',
        loss_weight=0.1,
    ))

# optimizer
optim_wrapper = dict(
    constructor='DefaultOptimWrapperConstructor',
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=0.00005))

# learning policy
# Fixed



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule, build_norm_layer
from mmengine.model import constant_init

from mmedit.registry import MODELS


@MODELS.register_module()
class WGANNoiseTo2DFeat(nn.Module):
    """Module used in WGAN-GP to transform 1D noise tensor in order [N, C] to
    2D shape feature tensor in order [N, C, H, W].

    Args:
        noise_size (int): Size of the input noise vector.
        out_channels (int): The channel number of the output feature.
        act_cfg (dict, optional): Config for the activation layer. Defaults to
            dict(type='ReLU').
        norm_cfg (dict, optional): Config dict to build norm layer. Defaults to
            dict(type='BN').
        order (tuple, optional): The order of conv/norm/activation layers. It
            is a sequence of "conv", "norm" and "act". Common examples are
            ("conv", "norm", "act") and ("act", "conv", "norm"). Defaults to
            ('linear', 'act', 'norm').
    """

    def __init__(self,
                 noise_size,
                 out_channels,
                 act_cfg=dict(type='ReLU'),
                 norm_cfg=dict(type='BN'),
                 order=('linear', 'act', 'norm')):
        super().__init__()
        self.noise_size = noise_size
        self.out_channels = out_channels
        self.with_activation = act_cfg is not None
        self.with_norm = norm_cfg is not None
        self.order = order
        assert len(order) == 3 and set(order) == set(['linear', 'act', 'norm'])

        # w/o bias, because the bias is added after reshaping the tensor to
        # 2D feature
        self.linear = nn.Linear(noise_size, out_channels * 16, bias=False)

        if self.with_activation:
            self.activation = MODELS.build(act_cfg)

        # add bias for reshaped 2D feature.
        self.register_parameter(
            'bias', nn.Parameter(torch.zeros(1, out_channels, 1, 1)))

        if self.with_norm:
            _, self.norm = build_norm_layer(norm_cfg, out_channels)
        self._init_weight()

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input noise tensor with shape (n, c).

        Returns:
            Tensor: Forward results with shape (n, c, 4, 4).
        """
        assert x.ndim == 2
        for order in self.order:
            if order == 'linear':
                x = self.linear(x)
                # [n, c, 4, 4]
                x = torch.reshape(x, (-1, self.out_channels, 4, 4))
                x = x + self.bias
            elif order == 'act' and self.with_activation:
                x = self.activation(x)
            elif order == 'norm' and self.with_norm:
                x = self.norm(x)

        return x

    def _init_weight(self):
        """Initialize weights for the model."""
        nn.init.normal_(self.linear.weight, 0., 1.)
        if self.bias is not None:
            nn.init.constant_(self.bias, 0.)
        if self.with_norm:
            constant_init(self.norm, 1, bias=0)


class WGANDecisionHead(nn.Module):
    """Module used in WGAN-GP to get the final prediction result with 4x4
    resolution input tensor in the bottom of the discriminator.

    Args:
        in_channels (int): Number of channels in input feature map.
        mid_channels (int): Number of channels in feature map after
            convolution.
        out_channels (int): The channel number of the final output layer.
        bias (bool, optional): Whether to use bias parameter. Defaults to True.
        act_cfg (dict, optional): Config for the activation layer. Defaults to
            dict(type='ReLU').
        out_act (dict, optional): Config for the activation layer of output
            layer. Defaults to None.
        norm_cfg (dict, optional): Config dict to build norm layer. Defaults to
            dict(type='LN2d').
    """

    def __init__(self,
                 in_channels,
                 mid_channels,
                 out_channels,
                 bias=True,
                 act_cfg=dict(type='ReLU'),
                 out_act=None,
                 norm_cfg=dict(type='LN2d')):
        super().__init__()
        self.in_channels = in_channels
        self.mid_channels = mid_channels
        self.out_channels = out_channels
        self.with_out_activation = out_act is not None

        # setup conv layer
        self.conv = ConvLNModule(
            in_channels,
            feature_shape=(mid_channels, 1, 1),
            kernel_size=4,
            out_channels=mid_channels,
            act_cfg=act_cfg,
            norm_cfg=norm_cfg,
            order=('conv', 'norm', 'act'))
        # setup linear layer
        self.linear = nn.Linear(
            self.mid_channels, self.out_channels, bias=bias)

        if self.with_out_activation:
            self.out_activation = MODELS.build(out_act)

        self._init_weight()

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """
        x = self.conv(x)
        x = torch.reshape(x, (x.shape[0], -1))
        x = self.linear(x)
        if self.with_out_activation:
            x = self.out_activation(x)
        return x

    def _init_weight(self):
        """Initialize weights for the model."""
        nn.init.normal_(self.linear.weight, 0., 1.)
        nn.init.constant_(self.linear.bias, 0.)


@MODELS.register_module()
class ConvLNModule(ConvModule):
    r"""ConvModule with Layer Normalization.

    In this module, we inherit default ``mmcv.cnn.ConvModule`` and deal with
    the situation that 'norm_cfg' is 'LN2d' or 'GN'. We adopt 'GN' as a
    replacement for layer normalization referring to:
    https://github.com/LynnHo/DCGAN-LSGAN-WGAN-GP-DRAGAN-Pytorch/blob/master/module.py # noqa

    Args:
        feature_shape (tuple): The shape of feature map that will be.
    """

    def __init__(self, *args, feature_shape=None, **kwargs):
        if 'norm_cfg' in kwargs and kwargs['norm_cfg'] is not None and kwargs[
                'norm_cfg']['type'] in ['LN2d', 'GN']:
            nkwargs = deepcopy(kwargs)
            nkwargs['norm_cfg'] = None
            super().__init__(*args, **nkwargs)
            self.with_norm = True
            self.norm_name = kwargs['norm_cfg']['type']
            if self.norm_name == 'LN2d':
                norm = nn.LayerNorm(feature_shape)
                self.add_module(self.norm_name, norm)
            else:
                norm = nn.GroupNorm(1, feature_shape[0])
                self.add_module(self.norm_name, norm)
        else:
            super().__init__(*args, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
import os.path as osp
from typing import Dict, List, Optional, Union

import torch
from mmengine.logging import MMLogger
from mmengine.model import BaseModel
from mmengine.runner import set_random_seed
from mmengine.runner.checkpoint import _load_checkpoint
from tqdm.auto import tqdm

from mmedit.registry import DIFFUSION_SCHEDULERS, MODELS
from .clip_wrapper import load_clip_submodels
from .vae import AutoencoderKL

logger = MMLogger.get_current_instance()


@MODELS.register_module('sd')
@MODELS.register_module()
class StableDiffusion(BaseModel):
    """class to run stable diffsuion pipeline.

    Args:
        diffusion_scheduler(dict): Diffusion scheduler config.
        unet_cfg(dict): Unet config.
        vae_cfg(dict): Vae config.
        pretrained_ckpt_path(dict):
            Pretrained ckpt path for submodels in stable diffusion.
        requires_safety_checker(bool):
            whether to run safety checker after image generated.
        unet_sample_size(int): sampel size for unet.
    """

    def __init__(self,
                 diffusion_scheduler,
                 unet,
                 vae,
                 requires_safety_checker=True,
                 unet_sample_size=64,
                 init_cfg=None):
        super().__init__()

        self.device = torch.device('cpu')
        self.submodels = [
            'tokenizer', 'vae', 'scheduler', 'unet', 'feature_extractor',
            'text_encoder'
        ]
        self.requires_safety_checker = requires_safety_checker

        self.scheduler = DIFFUSION_SCHEDULERS.build(
            diffusion_scheduler) if isinstance(diffusion_scheduler,
                                               dict) else diffusion_scheduler
        self.scheduler.order = 1
        self.scheduler.init_noise_sigma = 1.0

        self.unet_sample_size = unet_sample_size
        self.unet = MODELS.build(unet) if isinstance(unet, dict) else unet

        self.vae = AutoencoderKL(**vae) if isinstance(vae, dict) else vae
        self.vae_scale_factor = 2**(len(self.vae.block_out_channels) - 1)

        self.init_cfg = init_cfg
        self.init_weights()

    def init_weights(self):
        """load pretrained ckpt for each submodel."""
        if self.init_cfg is not None and self.init_cfg['type'] == 'Pretrained':
            map_location = self.init_cfg.get('map_location', 'cpu')
            pretrained_model_path = self.init_cfg.get('pretrained_model_path',
                                                      None)
            if pretrained_model_path:
                unet_ckpt_path = osp.join(pretrained_model_path, 'unet',
                                          'diffusion_pytorch_model.bin')
                if unet_ckpt_path:
                    state_dict = _load_checkpoint(unet_ckpt_path, map_location)
                    self.unet.load_state_dict(state_dict, strict=True)

                vae_ckpt_path = osp.join(pretrained_model_path, 'vae',
                                         'diffusion_pytorch_model.bin')
                if vae_ckpt_path:
                    state_dict = _load_checkpoint(vae_ckpt_path, map_location)
                    self.vae.load_state_dict(state_dict, strict=True)

        self.tokenizer, self.feature_extractor, self.text_encoder, self.safety_checker = load_clip_submodels(  # noqa
            self.init_cfg, self.submodels, self.requires_safety_checker)

    def to(self, torch_device: Optional[Union[str, torch.device]] = None):
        """put submodels to torch device.

        Args:
            torch_device(Optional[Union[str, torch.device]]):
                device to put, default to None.

        Returns:
            self(StableDiffusion):
                class instance itsself.
        """
        if torch_device is None:
            return self

        for name in self.submodels:
            module = getattr(self, name)
            if isinstance(module, torch.nn.Module):
                module.to(torch_device)
        self.device = torch.device(torch_device)
        return self

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]],
              height: Optional[int] = None,
              width: Optional[int] = None,
              num_inference_steps: int = 50,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_images_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[torch.Generator] = None,
              latents: Optional[torch.FloatTensor] = None,
              show_progress=True,
              seed=1):
        """Function invoked when calling the pipeline for generation.

        Args:
            prompt (`str` or `List[str]`):
                The prompt or prompts to guide the image generation.
            height (`int`, *optional*,
                defaults to self.unet_sample_size * self.vae_scale_factor):
                The height in pixels of the generated image.
            width (`int`, *optional*,
                defaults to self.unet_sample_size * self.vae_scale_factor):
                The width in pixels of the generated image.
            num_inference_steps (`int`, *optional*, defaults to 50):
                The number of denoising steps.
                More denoising steps usually lead to a higher
                quality image at the expense of slower inference.
            guidance_scale (`float`, *optional*, defaults to 7.5):
                Guidance scale as defined in
                [Classifier-Free Diffusion Guidance]
                (https://arxiv.org/abs/2207.12598).
            negative_prompt (`str` or `List[str]`, *optional*):
                The prompt or prompts not to guide the image generation.
                Ignored when not using guidance (i.e., ignored
                if `guidance_scale` is less than `1`).
            num_images_per_prompt (`int`, *optional*, defaults to 1):
                The number of images to generate per prompt.
            eta (`float`, *optional*, defaults to 0.0):
                Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                [`schedulers.DDIMScheduler`], will be ignored for others.
            generator (`torch.Generator`, *optional*):
                A [torch generator] to make generation deterministic.
            latents (`torch.FloatTensor`, *optional*):
                Pre-generated noisy latents,
                sampled from a Gaussian distribution,
                to be used as inputs for image generation.
                Can be used to tweak the same generation
                with different prompts.
                If not provided, a latents tensor will be
                generated by sampling using the supplied random `generator`.

        Returns:
            dict:['samples', 'nsfw_content_detected']:
                'samples': image result samples
                'nsfw_content_detected': nsfw content flags for image samples.
        """
        set_random_seed(seed=seed)

        # 0. Default height and width to unet
        height = height or self.unet_sample_size * self.vae_scale_factor
        width = width or self.unet_sample_size * self.vae_scale_factor

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(prompt, height, width)

        # 2. Define call parameters
        batch_size = 1 if isinstance(prompt, str) else len(prompt)
        device = self.device
        # here `guidance_scale` is defined analog to the
        # guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        # 3. Encode input prompt
        text_embeddings = self._encode_prompt(prompt, device,
                                              num_images_per_prompt,
                                              do_classifier_free_guidance,
                                              negative_prompt)

        # 4. Prepare timesteps
        # self.scheduler.set_timesteps(num_inference_steps, device=device)
        self.scheduler.set_timesteps(num_inference_steps)
        timesteps = self.scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.in_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            text_embeddings.dtype,
            device,
            generator,
            latents,
        )

        # 6. Prepare extra step kwargs.
        # TODO: Logic should ideally just be moved out of the pipeline
        extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

        # 7. Denoising loop
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat(
                [latents] * 2) if do_classifier_free_guidance else latents
            # latent_model_input = \
            # self.scheduler.scale_model_input(latent_model_input, t)

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input, t,
                encoder_hidden_states=text_embeddings)['outputs']

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.scheduler.step(
                    noise_pred, t, latents, **extra_step_kwargs)['prev_sample']

        # 8. Post-processing
        image = self.decode_latents(latents)

        # 9. Run safety checker
        image, has_nsfw_concept = self.run_safety_checker(
            image, device, text_embeddings.dtype)
        image = image[0].permute([2, 0, 1])

        return {'samples': image, 'nsfw_content_detected': has_nsfw_concept}

    def _encode_prompt(self, prompt, device, num_images_per_prompt,
                       do_classifier_free_guidance, negative_prompt):
        """Encodes the prompt into text encoder hidden states.

        Args:
            prompt (str or list(int)):
                prompt to be encoded.
            device: (torch.device):
                torch device.
            num_images_per_prompt (int):
                number of images that should be generated per prompt.
            do_classifier_free_guidance (`bool`):
                whether to use classifier free guidance or not.
            negative_prompt (str or List[str]):
                The prompt or prompts not to guide the image generation.
                Ignored when not using guidance (i.e., ignored
                if `guidance_scale` is less than `1`).

        Returns:
            text_embeddings (torch.Tensor):
                text embeddings generated by clip text encoder.
        """
        batch_size = len(prompt) if isinstance(prompt, list) else 1

        text_inputs = self.tokenizer(
            prompt,
            padding='max_length',
            max_length=self.tokenizer.model_max_length,
            truncation=True,
            return_tensors='pt',
        )
        text_input_ids = text_inputs.input_ids
        untruncated_ids = self.tokenizer(
            prompt, padding='max_length', return_tensors='pt').input_ids

        if not torch.equal(text_input_ids, untruncated_ids):
            removed_text = self.tokenizer.batch_decode(
                untruncated_ids[:, self.tokenizer.model_max_length - 1:-1])
            logger.warning(
                'The following part of your input was truncated because CLIP'
                ' can only handle sequences up to'
                f' {self.tokenizer.model_max_length} tokens: {removed_text}')

        if hasattr(self.text_encoder.config, 'use_attention_mask'
                   ) and self.text_encoder.config.use_attention_mask:
            attention_mask = text_inputs.attention_mask.to(device)
        else:
            attention_mask = None

        text_embeddings = self.text_encoder(
            text_input_ids.to(device),
            attention_mask=attention_mask,
        )
        text_embeddings = text_embeddings[0]

        # duplicate text embeddings for each generation per prompt,
        bs_embed, seq_len, _ = text_embeddings.shape
        text_embeddings = text_embeddings.repeat(1, num_images_per_prompt, 1)
        text_embeddings = text_embeddings.view(
            bs_embed * num_images_per_prompt, seq_len, -1)

        # get unconditional embeddings for classifier free guidance
        if do_classifier_free_guidance:
            uncond_tokens: List[str]
            if negative_prompt is None:
                uncond_tokens = [''] * batch_size
            elif type(prompt) is not type(negative_prompt):
                raise TypeError(
                    f'`negative_prompt` should be the same type to `prompt`,'
                    f'but got {type(negative_prompt)} !='
                    f' {type(prompt)}.')
            elif isinstance(negative_prompt, str):
                uncond_tokens = [negative_prompt]
            elif batch_size != len(negative_prompt):
                raise ValueError(
                    f'`negative_prompt`: {negative_prompt} has '
                    f'batch size {len(negative_prompt)}, but `prompt`:'
                    f' {prompt} has batch size {batch_size}.'
                    f' Please make sure that passed `negative_prompt` matches'
                    ' the batch size of `prompt`.')
            else:
                uncond_tokens = negative_prompt

            max_length = text_input_ids.shape[-1]
            uncond_input = self.tokenizer(
                uncond_tokens,
                padding='max_length',
                max_length=max_length,
                truncation=True,
                return_tensors='pt',
            )

            if hasattr(self.text_encoder.config, 'use_attention_mask'
                       ) and self.text_encoder.config.use_attention_mask:
                attention_mask = uncond_input.attention_mask.to(device)
            else:
                attention_mask = None

            uncond_embeddings = self.text_encoder(
                uncond_input.input_ids.to(device),
                attention_mask=attention_mask,
            )
            uncond_embeddings = uncond_embeddings[0]

            # duplicate unconditional embeddings for
            # each generation per prompt, using mps friendly method
            seq_len = uncond_embeddings.shape[1]
            uncond_embeddings = uncond_embeddings.repeat(
                1, num_images_per_prompt, 1)
            uncond_embeddings = uncond_embeddings.view(
                batch_size * num_images_per_prompt, seq_len, -1)

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional
            # and text embeddings into a single batch
            # to avoid doing two forward passes
            text_embeddings = torch.cat([uncond_embeddings, text_embeddings])

        return text_embeddings

    def run_safety_checker(self, image, device, dtype):
        """run safety checker to check whether image has nsfw content.

        Args:
            image (numpy.ndarray):
                image generated by stable diffusion.
            device (torch.device):
                device to run safety checker.
            dtype (torch.dtype):
                float type to run.

        Returns:
            image (numpy.ndarray):
                black image if nsfw content detected else input image.
            has_nsfw_concept (list[bool]):
                flag list to indicate nsfw content detected.
        """
        if self.safety_checker is not None:
            safety_checker_input = self.feature_extractor(
                image[0], return_tensors='pt').to(device)
            image, has_nsfw_concept = self.safety_checker(
                images=image,
                clip_input=safety_checker_input.pixel_values.to(dtype))
        else:
            has_nsfw_concept = None
        return image, has_nsfw_concept

    def decode_latents(self, latents):
        """use vae to decode latents.

        Args:
            latents (torch.Tensor): latents to decode.

        Returns:
            image (numpy.ndarray): image result.
        """
        latents = 1 / 0.18215 * latents
        image = self.vae.decode(latents).sample
        image = (image / 2 + 0.5).clamp(0, 1)
        # we always cast to float32 as this does not cause
        # significant overhead and is compatible with bfloa16
        image = image.cpu().permute(0, 2, 3, 1).float()
        return image

    def prepare_extra_step_kwargs(self, generator, eta):
        """prepare extra kwargs for the scheduler step.

        Args:
            generator (torch.Generator):
                generator for random functions.
            eta (float):
                eta (η) is only used with the DDIMScheduler,
                it will be ignored for other schedulers.
                eta corresponds to η in DDIM paper:
                https://arxiv.org/abs/2010.02502
                and should be between [0, 1]

        Return:
            extra_step_kwargs (dict):
                dict contains 'generator' and 'eta'
        """
        accepts_eta = 'eta' in set(
            inspect.signature(self.scheduler.step).parameters.keys())
        extra_step_kwargs = {}
        if accepts_eta:
            extra_step_kwargs['eta'] = eta

        # check if the scheduler accepts generator
        accepts_generator = 'generator' in set(
            inspect.signature(self.scheduler.step).parameters.keys())
        if accepts_generator:
            extra_step_kwargs['generator'] = generator
        return extra_step_kwargs

    def check_inputs(self, prompt, height, width):
        """check whether inputs are in suitable format or not."""

        if not isinstance(prompt, str) and not isinstance(prompt, list):
            raise ValueError(f'`prompt` has to be of '
                             f'type `str` or `list` but is {type(prompt)}')

        if height % 8 != 0 or width % 8 != 0:
            raise ValueError(f'`height` and `width` have to be divisible '
                             f'by 8 but are {height} and {width}.')

    def prepare_latents(self,
                        batch_size,
                        num_channels_latents,
                        height,
                        width,
                        dtype,
                        device,
                        generator,
                        latents=None):
        """prepare latents for diffusion to run in latent space.

        Args:
            batch_size (int): batch size.
            num_channels_latents (int): latent channel nums.
            height (int): image height.
            width (int): image width.
            dtype (torch.dtype): float type.
            device (torch.device): torch device.
            generator (torch.Generator):
                generator for random functions, defaults to None.
            latents (torch.Tensor):
                Pre-generated noisy latents, defaults to None.

        Return:
            latents (torch.Tensor): prepared latents.
        """
        shape = (batch_size, num_channels_latents,
                 height // self.vae_scale_factor,
                 width // self.vae_scale_factor)
        if latents is None:
            latents = torch.randn(
                shape, generator=generator, device=device, dtype=dtype)
        else:
            if latents.shape != shape:
                raise ValueError(f'Unexpected latents shape, '
                                 f'got {latents.shape}, expected {shape}')
            latents = latents.to(device)

        # scale the initial noise by the standard
        # deviation required by the scheduler
        latents = latents * self.scheduler.init_noise_sigma
        return latents

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[list] = None,
                mode: str = 'tensor') -> Union[Dict[str, torch.Tensor], list]:
        """forward is not implemented now."""
        raise NotImplementedError(
            'Forward is not implemented now, please use infer.')



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Dict, Union

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from mmengine import MessageHub
from mmengine.logging import MMLogger
from mmengine.model import is_model_wrapper
from mmengine.optim import OptimWrapper, OptimWrapperDict
from torch import Tensor

from mmedit.registry import MODELS
from mmedit.structures import EditDataSample
from ...utils import set_requires_grad
from ..stylegan2 import StyleGAN2

ModelType = Union[Dict, nn.Module]
TrainInput = Union[dict, Tensor]


@MODELS.register_module()
class MSPIEStyleGAN2(StyleGAN2):
    """MS-PIE StyleGAN2.

    In this GAN, we adopt the MS-PIE training schedule so that multi-scale
    images can be generated with a single generator. Details can be found in:
    Positional Encoding as Spatial Inductive Bias in GANs, CVPR2021.

    Args:
        train_settings (dict): Config for training settings.
            Defaults to `dict()`.
    """

    def __init__(self, *args, train_settings=dict(), **kwargs):
        super().__init__(*args, **kwargs)
        self.train_settings = deepcopy(train_settings)
        # set the number of upsampling blocks. This value will be used to
        # calculate the current result size according to the size of the input
        # feature map, e.g., positional encoding map
        self.num_upblocks = self.train_settings.get('num_upblocks', 6)

        # multiple input scales (a list of int) that will be added to the
        # original starting scale.
        self.multi_input_scales = self.train_settings.get('multi_input_scales')
        self.multi_scale_probability = self.train_settings.get(
            'multi_scale_probability')

    def train_step(self, data: dict,
                   optim_wrapper: OptimWrapperDict) -> Dict[str, Tensor]:
        """Train GAN model. In the training of GAN models, generator and
        discriminator are updated alternatively. In MMEditing's design,
        `self.train_step` is called with data input. Therefore we always update
        discriminator, whose updating is relay on real data, and then determine
        if the generator needs to be updated based on the current number of
        iterations. More details about whether to update generator can be found
        in :meth:`should_gen_update`.

        Args:
            data (dict): Data sampled from dataloader.
            optim_wrapper (OptimWrapperDict): OptimWrapperDict instance
                contains OptimWrapper of generator and discriminator.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """
        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')
        data = self.data_preprocessor(data, True)

        disc_optimizer_wrapper: OptimWrapper = optim_wrapper['discriminator']
        disc_accu_iters = disc_optimizer_wrapper._accumulative_counts

        with disc_optimizer_wrapper.optim_context(self.discriminator):
            log_vars = self.train_discriminator(
                **data, optimizer_wrapper=disc_optimizer_wrapper)

        # add 1 to `curr_iter` because iter is updated in train loop.
        # Whether to update the generator. We update generator with
        # discriminator is fully updated for `self.n_discriminator_steps`
        # iterations. And one full updating for discriminator contains
        # `disc_accu_counts` times of grad accumulations.
        if (curr_iter + 1) % (self.discriminator_steps * disc_accu_iters) == 0:
            set_requires_grad(self.discriminator, False)
            gen_optimizer_wrapper = optim_wrapper['generator']
            gen_accu_iters = gen_optimizer_wrapper._accumulative_counts

            log_vars_gen_list = []
            # init optimizer wrapper status for generator manually
            gen_optimizer_wrapper.initialize_count_status(
                self.generator, 0, self.generator_steps * gen_accu_iters)
            for _ in range(self.generator_steps * gen_accu_iters):
                with gen_optimizer_wrapper.optim_context(self.generator):
                    log_vars_gen = self.train_generator(
                        **data, optimizer_wrapper=gen_optimizer_wrapper)

                log_vars_gen_list.append(log_vars_gen)
            log_vars_gen = self.gather_log_vars(log_vars_gen_list)
            log_vars_gen.pop('loss', None)  # remove 'loss' from gen logs

            set_requires_grad(self.discriminator, True)

            # only do ema after generator update
            if self.with_ema_gen and (curr_iter + 1) >= (
                    self.ema_start * self.discriminator_steps *
                    disc_accu_iters):
                self.generator_ema.update_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)
                # if not update buffer, copy buffer from orig model
                if not self.generator_ema.update_buffers:
                    self.generator_ema.sync_buffers(
                        self.generator.module if is_model_wrapper(
                            self.generator) else self.generator)
            elif self.with_ema_gen:
                # before ema, copy weights from orig
                self.generator_ema.sync_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)

            log_vars.update(log_vars_gen)

        return log_vars

    def train_generator(self, inputs: dict, data_samples: EditDataSample,
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (TrainInput): Inputs from dataloader.
            data_samples (EditDataSample): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        num_batches = len(data_samples)

        noise = self.noise_fn(num_batches=num_batches)
        fake_imgs = self.generator(
            noise, return_noise=False, chosen_scale=self.chosen_scale)

        disc_pred_fake = self.discriminator(fake_imgs)
        parsed_loss, log_vars = self.gen_loss(disc_pred_fake, num_batches)

        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def train_discriminator(self, inputs: dict, data_samples: EditDataSample,
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (TrainInput): Inputs from dataloader.
            data_samples (EditDataSample): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.
        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs = data_samples.gt_img

        if dist.is_initialized():
            # randomly sample a scale for current training iteration
            chosen_scale = np.random.choice(self.multi_input_scales, 1,
                                            self.multi_scale_probability)[0]

            chosen_scale = torch.tensor(chosen_scale, dtype=torch.int).cuda()
            dist.broadcast(chosen_scale, 0)
            chosen_scale = int(chosen_scale.item())

        else:
            logger = MMLogger.get_current_instance()
            logger.info(
                'Distributed training has not been initialized. Degrade to '
                'the standard stylegan2')
            chosen_scale = 0

        curr_size = (4 + chosen_scale) * (2**self.num_upblocks)
        # adjust the shape of images
        if real_imgs.shape[-2:] != (curr_size, curr_size):
            real_imgs = F.interpolate(
                real_imgs,
                size=(curr_size, curr_size),
                mode='bilinear',
                align_corners=True)

        num_batches = real_imgs.shape[0]

        noise_batch = self.noise_fn(num_batches=num_batches)
        with torch.no_grad():
            fake_imgs = self.generator(
                noise_batch, return_noise=False, chosen_scale=chosen_scale)
        # store chosen scale for training generator
        setattr(self, 'chosen_scale', chosen_scale)

        disc_pred_fake = self.discriminator(fake_imgs)
        disc_pred_real = self.discriminator(real_imgs)

        parsed_losses, log_vars = self.disc_loss(disc_pred_fake,
                                                 disc_pred_real, real_imgs)
        optimizer_wrapper.update_params(parsed_losses)
        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch
from mmengine.config import Config
from mmengine.model import BaseModel
from mmengine.optim import OptimWrapperDict

from mmedit.registry import MODELS
from mmedit.structures import EditDataSample
from .color_utils import get_colorization_data, lab2rgb


@MODELS.register_module()
class InstColorization(BaseModel):
    """Colorization InstColorization  method.

    This Colorization is implemented according to the paper:
        Instance-aware Image Colorization, CVPR 2020

    Adapted from 'https://github.com/ericsujw/InstColorization.git'
    'InstColorization/models/train_model'
    Copyright (c) 2020, Su, under MIT License.

    Args:
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`.
        image_model (dict): Config for single image model
        instance_model (dict): Config for instance model
        fusion_model (dict): Config for fusion model
        color_data_opt (dict): Option for colorspace conversion
        which_direction (str): AtoB or BtoA
        loss (dict): Config for loss.
        init_cfg  (str): Initialization config dict. Default: None.
        train_cfg (dict): Config for training. Default: None.
        test_cfg (dict): Config for testing. Default: None.
    """

    def __init__(self,
                 data_preprocessor: Union[dict, Config],
                 image_model,
                 instance_model,
                 fusion_model,
                 color_data_opt,
                 which_direction='AtoB',
                 loss=None,
                 init_cfg=None,
                 train_cfg=None,
                 test_cfg=None):

        super().__init__(
            init_cfg=init_cfg, data_preprocessor=data_preprocessor)

        # colorization networks
        # image_model: used to colorize a single image
        self.image_model = MODELS.build(image_model)

        # instance model: used to colorize cropped instance
        self.instance_model = MODELS.build(instance_model)

        # fusion model: input a single image with related instance features
        self.fusion_model = MODELS.build(fusion_model)

        self.color_data_opt = color_data_opt
        self.which_direction = which_direction

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[List[EditDataSample]] = None,
                mode: str = 'tensor',
                **kwargs):
        """Returns losses or predictions of training, validation, testing, and
        simple inference process.

        ``forward`` method of BaseModel is an abstract method, its subclasses
        must implement this method.

        Accepts ``inputs`` and ``data_samples`` processed by
        :attr:`data_preprocessor`, and returns results according to mode
        arguments.

        During non-distributed training, validation, and testing process,
        ``forward`` will be called by ``BaseModel.train_step``,
        ``BaseModel.val_step`` and ``BaseModel.val_step`` directly.

        During distributed data parallel training process,
        ``MMSeparateDistributedDataParallel.train_step`` will first call
        ``DistributedDataParallel.forward`` to enable automatic
        gradient synchronization, and then call ``forward`` to get training
        loss.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.
            mode (str): mode should be one of ``loss``, ``predict`` and
                ``tensor``. Default: 'tensor'.

                - ``loss``: Called by ``train_step`` and return loss ``dict``
                  used for logging
                - ``predict``: Called by ``val_step`` and ``test_step``
                  and return list of ``BaseDataElement`` results used for
                  computing metric.
                - ``tensor``: Called by custom use to get ``Tensor`` type
                  results.

        Returns:
            ForwardResults:

                - If ``mode == loss``, return a ``dict`` of loss tensor used
                  for backward and logging.
                - If ``mode == predict``, return a ``list`` of
                  :obj:`BaseDataElement` for computing metric
                  and getting inference result.
                - If ``mode == tensor``, return a tensor or ``tuple`` of tensor
                  or ``dict`` or tensor for custom use.
        """

        if mode == 'tensor':
            return self.forward_tensor(inputs, data_samples, **kwargs)

        elif mode == 'predict':
            predictions = self.forward_inference(inputs, data_samples,
                                                 **kwargs)
            predictions = self.convert_to_datasample(data_samples, predictions)
            return predictions

        elif mode == 'loss':
            return self.forward_train(inputs, data_samples, **kwargs)

    def convert_to_datasample(self, inputs, data_samples):
        """Add predictions and destructed inputs (if passed) to data samples.

        Args:
            inputs (Optional[torch.Tensor]): The input of model. Defaults to
                None.
            data_samples (List[EditDataSample]): The data samples loaded from
                dataloader.

        Returns:
            List[EditDataSample]: Modified data samples.
        """

        for data_sample, output in zip(inputs, data_samples):
            data_sample.output = output
        return inputs

    def forward_train(self, inputs, data_samples=None, **kwargs):
        """Forward function for training."""
        raise NotImplementedError(
            'Instance Colorization has not supported training.')

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapperDict) -> Dict[str, torch.Tensor]:
        """Train step function.

        Args:
            data (List[dict]): Batch of data as input.
            optim_wrapper (dict[torch.optim.Optimizer]): Dict with optimizers
                for generator and discriminator (if have).
        Returns:
            dict: Dict with loss, information for logger, the number of
                samples and results for visualization.
        """
        raise NotImplementedError(
            'Instance Colorization has not supported training.')

    def forward_inference(self, inputs, data_samples=None, **kwargs):
        """Forward inference. Returns predictions of validation, testing.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.

        Returns:
            List[EditDataSample]: predictions.
        """
        feats = self.forward_tensor(inputs, data_samples, **kwargs)
        feats = self.data_preprocessor.destruct(feats, data_samples)
        predictions = []
        for idx in range(feats.shape[0]):
            pred_img = feats[idx].to('cpu')
            predictions.append(
                EditDataSample(
                    pred_img=pred_img, metainfo=data_samples[idx].metainfo))

        return predictions

    def forward_tensor(self, inputs, data_samples):
        """Forward function in tensor mode.

        Args:
            inputs (torch.Tensor): Input tensor.
            data_sample (dict): Dict contains data sample.

        Returns:
            dict: Dict contains output results.
        """

        #  prepare data

        assert len(data_samples) == 1, \
            'fusion model supports only one image due to different numbers '\
            'of instances of different images'

        full_img_data = get_colorization_data(inputs, self.color_data_opt)
        AtoB = self.which_direction == 'AtoB'

        # preprocess input for a single image
        full_real_A = full_img_data['A' if AtoB else 'B']
        full_hint_B = full_img_data['hint_B']
        full_mask_B = full_img_data['mask_B']

        if not data_samples.empty_box[0]:
            # preprocess instance input
            cropped_img = data_samples.cropped_img[0]
            box_info_list = [
                data_samples.box_info[0], data_samples.box_info_2x[0],
                data_samples.box_info_4x[0], data_samples.box_info_8x[0]
            ]
            cropped_data = get_colorization_data(cropped_img,
                                                 self.color_data_opt)

            real_A = cropped_data['A' if AtoB else 'B']
            hint_B = cropped_data['hint_B']
            mask_B = cropped_data['mask_B']

            # network forward
            _, output, feature_map = self.instance_model(
                real_A, hint_B, mask_B)
            output = self.fusion_model(full_real_A, full_hint_B, full_mask_B,
                                       feature_map, box_info_list)

        else:
            _, output, _ = self.image_model(full_real_A, full_hint_B,
                                            full_mask_B)

        output = [
            full_real_A.type(torch.cuda.FloatTensor),
            output.type(torch.cuda.FloatTensor)
        ]
        output = torch.cat(output, dim=1)
        output = torch.clamp(lab2rgb(output, self.color_data_opt), 0.0, 1.0)
        return output
